/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: delay.cc,v 1.8 1994/01/12 22:57:01 bengt Exp $ */

/* Delays for use in Gnans systems */

/* Usage:
   (i)   as a C-DECLARATION, declare a pointer to a delay_line
   (ii)  AT TIME START, call delay_open, with the initial condition 
         (double) as argument
   (iii) AT TIME ACCEPTED, (not TIME t) store present value by calling
         delay_update with arguments the delay_line pointer and the quantity
         to be stored
   (iv)  The value of the delay_line tau units of time in the past can be read
         (without side-effects) with delay_read, arguments: the delay_line 
         and the delay
   (v)   AT TIME STOP, free allocated memory with delay_free.

   It appears as if the smart algorithms get confused by delays (actually,
   they where not designed for delay equations...), using a dumb algorithm
   might be a good idea (or restricting upper bound of step size).

   YES, this could be user friedlier, but then it had to be implemented
   with a preprocessor.

   This is a memory hog.
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include <stdlib.h>
#include "externals.h"
#include "gnanslib.h"		/* Contains the typedef of delay_line */
#include <errno.h>

static double *dmalloc(unsigned n);
static double *drealloc(double *old, unsigned n);

/****************************************************************/
/* This function returns a delay_line. Called at time START */
delay_line::delay_line()
{
  times = dmalloc(MALLOCUNIT);
  values = dmalloc(MALLOCUNIT);
  size = MALLOCUNIT;
  index = 0;
  last_read_index = 0;
}

double
delay_line::initialize(double init)
{
  initial = init;
  index = 0;
  last_read_index = 0;
  return init;
}

/* Returns the value of the delay_line d, delaytime time units ago */
double 
delay_line::value(double delaytime)
{
  if (size == 0)
    return initial;
  double then = systemtime - delaytime;
  if (delaytime <= 0) {
    gnans_error(non_fatal,
		"Delay must be positive\n"
		"(Crystal ball not implemented :-)");
    terminate(true);
  }
  if (then <= times[0]) /* Very old? */
    return initial;

  if (then >= times[last_read_index]) {
    /* This request is for a later time than the last request (Normal case) */
    for (unsigned int i = last_read_index; i < index; i++) {
      if ((times[i] <= then) && (times[i+1] >= then)) {
	/* Found the interval; interpolate */
	last_read_index = i;
	return(values[i] +
	       (then - times[i])/(times[i+1] - times[i]) *
	       (values[i+1] - values[i]));
      }
    }
    /* Did not find it, must be a later time than last stored value */
    last_read_index = index;
    return(values[index] +
	   (then - times[index])/(systemtime - times[index]) *
	   (systemtime - values[index]));
  } else {
    /* Search backwards */
    for (unsigned int i = last_read_index; i > 0; i--) {
      if ((times[i-1] <= then) && (times[i] >= then)) {
	/* Found the interval; interpolate */
	last_read_index = i - 1;
	return(values[i-1] +
	       (then - times[i-1])/(times[i] - times[i-1]) *
	       (values[i] - values[i-1]));
      }
    }
  }
  /* If I got here something is wrong */
  gnans_error(fatal, "Error in delay -- delay_update forgotten?");
  terminate(true);
  return initial;		/* Avoid compiler warnings... */
}

/*
 * This routine is to be called once for every accepted value.
 * Except for when using dumb integration routines, MUST NOT be called
 * at every function evaluation.
 */
double
delay_line::update(double value)
{
  if (index == size) { /* Full? */
    size += MALLOCUNIT;	/* Yes, malloc more */
    times = drealloc(times, size);
    values = drealloc(values, size);
  }
  times[index] = systemtime;
  values[index++] = value;
  return value;
}

/* Frees the storage allocated for a delay_line. Called at time STOP */
void 
delay_line::close()
{
}

delay_line::~delay_line()
{
  free(times);
  free(values);
}

/****************************************************************/
/*
 * Allocates doubles, with error control.
 */
static double *
dmalloc(unsigned n)
{
  double *ptr = (double *) malloc(n * sizeof(double));
  if (ptr == NULL) {
    gnans_error(fatal, "Cannot allocate %d more bytes.", n * sizeof(double));
    return(NULL);
  } else {
    return(ptr);
  }
}

/*
 * Like realloc, but for doubles and with error control. Equal to dmalloc
 * if called with NULL argument.
 */
static double *
drealloc(double *old, unsigned n)
{
  double *ptr;
  unsigned int i;
  if (malloc_debug)
    gnans_error(debug, "Realloc called with n = %d; old = %x",
                n, (long int) old);
  if (! old)
    return dmalloc(n);
  
  ptr = (double *) realloc(old, n * sizeof(double));

  if (ptr && malloc_debug)
    gnans_error(debug, "Realloc succeeded; new = %x", (long int) ptr);
  
  if (ptr == NULL) {
    if (errno == EINVAL)
      gnans_error(warning,
                  "Error EINVAL by realloc, doing it the primitive way");
    else if (errno == ENOMEM)
      gnans_error(warning,
                  "Error ENOMEM by realloc, doing it the primitive way");
    else
      gnans_error(warning,
                  "Cannot reallocate memory, doing it the primitive way");
    ptr = dmalloc(n);
    for (i = 0; i < n - MALLOCUNIT; i++)
      ptr[i] = old[i];
    free(old);
  }
  return(ptr);
}
