/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: histo.cc,v 1.10 1994/08/19 18:10:31 bengt Exp $ */

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "gnanslib.h"

void
histo::open(double xmin_, double xmax_, double ymin_, double ymax_,
	     int xbins_, int ybins_, bool include_faroff_,
	     char *matrixname_, char *filename_)
{
  if (xmax_ <= xmin_ || ymax_ <= ymin_) {
    gnans_error(non_fatal,
		"xmax and ymax must be larger than xmin and ymax, respectivelly");
    return;
  }

  dimension = 2;
  xmin = xmin_;
  xmax = xmax_;
  ymin = ymin_;
  ymax = ymax_;
  xbins = xbins_;
  ybins = ybins_;
  include_faroff = include_faroff_;
  matrixname = matrixname_;
  filename = filename_;
  data = new long int[xbins*ybins];
  for (unsigned int i = 0; i < xbins*ybins; i++)
    data[i] = 0;
}
  
histo::histo(double xmin_, double xmax_,
	     double ymin_, double ymax_,
	     int xbins_, int ybins_, bool include_faroff_,
	     char *matrixname_, char *filename_)
{
  open(xmin_, xmax_,  ymin_, ymax_, xbins_, ybins_, include_faroff_,
       matrixname_, filename_);
}

histo::histo()
{
  dimension = 0;
}

void
histo::open(double xmin_, double xmax_, int xbins_, bool include_faroff_,
	     char *matrixname_, char *filename_)
{
  if (xmax_ <= xmin_) {
    gnans_error(non_fatal, "xmax must be larger than xmin");
    return;
  }

  dimension = 1;
  xmin = xmin_;
  xmax = xmax_;
  xbins = xbins_;
  include_faroff = include_faroff_;
  matrixname = matrixname_;
  filename = filename_;
  data = new long int[xbins];
  for (unsigned int i = 0; i < xbins; i++)
    data[i] = 0;
}

histo::histo(double xmin_, double xmax_, int xbins_, bool include_faroff_,
	     char *matrixname_, char *filename_)
{
  open(xmin_, xmax_, xbins_, include_faroff_, matrixname_, filename_);
}

inline static void
clip(int &x, int a, int b) {
  x = x < a ? a : ( x > b ? b : x);
}

void
histo::update(double x)
{
  int i;
  bool discard = false;
  
  if (dimension != 1) {
    gnans_error(non_fatal, "Histogram incorrectly initialized");
    return;
  }

  i = (unsigned int) ((x - xmin)/(xmax - xmin) * xbins);
  if (include_faroff)
    clip(i, 0, xbins-1);
  else
    discard = i < 0 || unsigned(i) >= xbins;

  if (!discard)
    data[i]++;
}

void
histo::update(double x, double y)
{
  int i, j;
  bool discard = false;
  
  if (dimension != 2) {
    gnans_error(non_fatal, "Histogram incorrectly initialized");
    return;
  }

  i = (unsigned int) ((x - xmin)/(xmax - xmin) * xbins);
  j = (unsigned int) ((y - ymin)/(ymax - ymin) * ybins);
  if (include_faroff) {
    clip(i, 0, xbins-1);
    clip(j, 0, ybins-1);
  } else
    discard = i < 0 || unsigned(i) >= xbins || j < 0 || unsigned(j) >= ybins;

  if (!discard)
    data[i*ybins + j]++;
}

#ifdef  WORDS_BIGENDIAN
#define TYPE 1020      	// Sun etc, column-wise, 32 bit signed ints, !text
#else
#define TYPE 0020
#endif

class Fmatrix {
public:
  long type;			/* type */
  long mrows;			/* row dimension */
  long ncols;			/* column dimension */
  long imagf;			/* flag indicating imag part */
  long namlen;			/* name length (including NULL) */
};

bool
histo::matlab_write(FILE *f)
{
  Fmatrix x;
  int result;

  x.type = TYPE;
  x.mrows = (dimension == 1) ? 1 : ybins;
  x.ncols = xbins;
  x.imagf = false;
  x.namlen = matrixname.length() + 1;

  result = fwrite(&x, sizeof(Fmatrix), 1, f);
  if (result != EOF)
    result = fwrite(matrixname.c_str(), sizeof(char), (int) x.namlen, f);

  if (result != EOF)
    switch (dimension) {
    case 1:
      result = fwrite(data, sizeof(long int), xbins, f);
      break;
    case 2:
      result = fwrite(data, sizeof(long int), xbins*ybins, f);
      break;
    default:
      gnans_error(non_fatal, "Incorrectly initialized histogram");
      break;
    }
      
  return result != EOF;
}

bool
histo::ascii_write(FILE *f)
{
  unsigned int i;
  int result = 1;
  
  switch (dimension) {
  case 1:    
    for (i = 0; i < xbins; i++)
      result = fprintf(f, "%g %ld\n",
		       i*(xmax - xmin)/xbins + xmin,
		       data[i]);
    break;
  case 2:
    for (i = 0; i < xbins; i++)
      for (unsigned int j = 0; j < ybins; j++)
	result = fprintf(f, "%g %g %ld\n",
			 i*(xmax - xmin)/xbins + xmin,
			 j*(ymax - ymin)/ybins + ymin,
			 data[i*ybins + j]);
    break;
  default:
    gnans_error(non_fatal, "Incorrectly initialized histogram");
    break;
  }
  return result != EOF;
}

bool
histo::float_write(FILE */*f*/)
{
  gnans_error(non_fatal,
	      "I am not programmed to handle that store format for histograms");
  return false;
}

void
histo::close()
{
  bool success;
  FILE *f;

  if (!dimension) {
    gnans_error(non_fatal, "Histogram incorrectly initialized");
    return;
  }

  if (!(f = fopen(filename.c_str(), "w"))) {
    gnans_error(non_fatal, 
		"Could not open histogram file %s for writing",
		filename.c_str());
    return;
  }

  switch (store_format) {
  case 0:			// Matlab
    success = matlab_write(f);
    break;
  case 2:			// Ascii
    success = ascii_write(f);
    break;
  case 7:			// raw/float
    success = float_write(f);
    break;
  case 9:
    success = true;
    break;
  default:
    gnans_error(non_fatal,
		"I am not programmed to handle that store format for histograms");
    success = false;
    break;
  }

  if (!success)
    gnans_error(non_fatal, 
		"Error occurred during writing of histogram to disk");
  fclose(f);
  delete data;
}

double
histo::chi_2(double N)
{
  if (dimension != 1) {
    gnans_error(non_fatal,
		"I can do chi 2 only on one-dimensional histograms");
    return 0;
  }
  
  double sum = 0;
  for (unsigned int i = 0; i < xbins; i++)
    sum += sqr(data[i] - N/xbins);
  return sum/(N * 1/xbins);
}
