/*
 Copyright (C) 1994 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id:  $

#ifdef __GNUG__
#pragma implementation
#endif

#include <float.h>
#include "externals.h"
#include "utillib.h"
#include "Rksuite.h"

#ifdef FORTRAN_UNDERSCORES
#define setup setup_
#define ct ct_
#define reset reset_
#define d1mach d1mach_
#endif

#ifdef FORTRAN_UPPERCASE
#define setup SETUP
#define ct CT
#define reset RESET
#define d1mach D1MACH
#endif

// External declarations for the integrator
extern "C" double 
d1mach(int *i);

extern "C" void
setup(int *neq, double *tstart, double *ystart, double *tend,
      double *tol, double *thres, int *method, char *task,
      int *errass, double *hstart, double *work, int *lenwrk, int *mesage);

extern "C" void
ct(fp f, double *tnow, double *ynow, double *ypnow,
   double *work, int *cflag);

extern "C" int 
reset(double *tendnu);

const char *
RksuiteIntegrator::documentation() const
{
  return
    "                   RKSUITE Release 1.0  November 1991                "
    "\n"
    "\n"
    "                              written by                             "
    "\n"
    "\n"
    "      R.W. Brankin (*),  I. Gladwell,  and  L.F. Shampine            "
    "\n"
    "\n"
    "RKSUITE is a suite of codes based on Runge-Kutta formulas that\n" 
    "solves the initial value problem for a first order system of ordinary\n"
    "differential equations using Runge-Kutta methods. "
    "\n"
    "The authors would very much appreciate receiving notification of     \n"
    "errors and constructive criticism of the suite.                      \n"
    "\n"
    " (*)\n"
    "      Numerical Algorithms Group Ltd. \n"
    "      Wilkinson House \n"
    "      Jordan Hill Road \n"
    "      Oxford OX2 8DR \n"
    "      U.K. \n"
    "      email: richard@nag.co.uk \n"
    "             na.brankin@na-net.ornl.gov \n"
    "      International phone: + 44 865 511245 \n"
    "      International fax: + 44 865 310139 \n"
    "\n";
}

static integration_error_message Rksuite_errors[] = {
  { 2, "This code is being used inefficient because step size\n"
       "has been reduced drastically many times to get answers. "},
  { 3, "A considerable amount of work has been expended \n"
       "in the integration. More than 5000 calls of F. "},
  { 4, "It appears that the problem is stiff. Rksuite is inefficient. "},
  { 5, "It does not appear possible to achieve the accuracy 'relerror' \n"},
  { 6, "The global error assessment may not be reliable \n"
       "beyond the current integration time."},
  { 0, ""}
};

RksuiteIntegrator::RksuiteIntegrator()
{
  neqn = -1;			// Signals impossible condition, to restart
  work = NULL;
  wrkint = NULL;
  thres = NULL;
  ypnow = NULL;
}

static double
macheps()
{
  int junk = 3;
  return d1mach(&junk);
}

bool
RksuiteIntegrator::initialize_integrator()
{
  setup_times();
  if (continuous_time_barf() ||  barf_on_wieners())
    return false;
    
  if (int(method_flag) > 3 || int(method_flag)  < 1) {
    gnans_error(non_fatal, 
                "Unsupported method_flag (= %d)  in Rksuite. \n",
		int(method_flag));
    return false;
  }

  if (relerror > 0.01) {
    gnans_error(non_fatal, "relerror must not be larger than 0.01");
    return false;
  }

  if (relerror < 10*macheps()) {
    gnans_error(non_fatal,
		"relerror must not be smaller than %g", 10*macheps());
  }
  
  int oldmf = method;
  method = int(method_flag);

  if (unsigned(neqn) != current_system.number_of_states || oldmf != method) {
    neqn = current_system.number_of_states;
    lenwrk = method == 1 ? 10*neqn : method == 2 ? 14*neqn : 16*neqn;
    lenint =  6*neqn;

    delete [] work;
    work = new double[(unsigned) lenwrk];

    delete [] thres;
    thres = new double[(unsigned) neqn];

    delete [] ypnow;
    ypnow = new double[(unsigned) neqn];

    delete [] wrkint;
    wrkint = new double[(unsigned) lenint];
  }

  for (int i = 0; i < neqn; i++)
    thres[i] = abserror/relerror;

  cflag = 1;
  double hstart = 0;		// select automatically first step-size
  int errass = 0;		// no global error assessment
  int mesage = double(algorithm_parameter_1) > 0; // no messages

  initialize_integration();
  double t_end = signed_step == 0 ? stoptime : starttime+signed_step;
  setup(&neqn, starttime, current_system.state_vector, &t_end, relerror,
        thres, &method, "c", &errass, &hstart, work, &lenwrk, &mesage);
  (*current_system.f_ptr)(starttime, current_system.state_vector, ypnow);

  return true;
}

// Define inline function to guarantee consistency
//inline 
bool
RksuiteIntegrator::call_integrator()
{
  // I find the "soft failure mechanism" obnoxious.
  // Keep calling if soft failure occurs. See doc for Rksuite.
  do {
    if (cflag <= 4)
      ct(current_system.f_ptr, 
	 systemtime,
	 current_system.state_vector, 
	 ypnow,
	 work,
	 &cflag);
    if (cflag > 1 && double(algorithm_parameter_2) > 0)
      gnans_error(debug, "cflag = %d, t = %g", cflag, double(systemtime));
  } while (cflag <= 4 && cflag > 1);

  return cflag == 1;
}

// The driver for Rksuite. The structure is similar to that of the Euler
// driver. Check the comments in the Euler driver.
bool
RksuiteIntegrator::loop(bool initialize)
{
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }

  if (storeincrement == 0.0) {
    /***************** Store everything *****************/
    if (signed_step != 0) {
      /***************** Prescribed stepsize ****************/
      double tend;
      do {
	store_stuff();
	  
	tend = systemtime + signed_step;
	if (systemtime != starttime)
	  reset(&tend);
	
        do 
	  call_integrator();
	while ((signed_step > 0 ? systemtime < tend : systemtime > tend)
	       && cflag == 1);
	if (cflag == 1)
	  (*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && cflag == 1);
    } else {
      /***************** Free stepsize ****************/
      do {
	store_stuff();
	call_integrator();
	
	if (cflag == 1)
	  (*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && cflag == 1);
    }
  } else {
    /***************** Store selectively *****************/
    if (signed_step != 0) {
      /***************** Prescribed stepsize ****************/
      double tend;
      do {
	store_if_necessary();
	  
	tend = systemtime + signed_step;
	if (systemtime != starttime)
	  reset(&tend);
	
        do 
	  call_integrator();
	while ((signed_step > 0 ? systemtime < tend : systemtime > tend)
	       && cflag == 1);
	if (cflag == 1)
	  (*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && cflag == 1);
    } else {
      /***************** Free stepsize ****************/
      do {
	store_if_necessary();
	call_integrator();
	
	if (cflag == 1)
	  (*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && cflag == 1);
    }
  }
  
  if (finished()) {
    if (!signed_step)
      store_stuff();
  }

  if (cflag > 4)
    integration_error(starttime, systemtime, cflag, Rksuite_errors);
  return finished();
}
