/* -*- c++ -*-
Copyright  1992, 1993 Bengt Mrtensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: Vect.h,v 1.2 1994/08/19 18:38:31 bengt Exp $ */

#ifndef _Vect_h
#define _Vect_h

#ifdef __GNUG__
#pragma interface
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef GNU_CONST
#ifdef __GNUG__
#define GNU_CONST const
#else
#define GNU_CONST
#endif
#endif

#ifdef __STRICT_ANSI__
#define _G_NO_NRV
#endif

#include <math.h>		// For sqrt

class Vect
{
  double *base;
  void barf(char *);
  unsigned int size;
public:
  Vect(unsigned int n);
  Vect(const Vect &v);
  Vect()				{ base = 0; size = 0; }
  ~Vect()				{ delete [] base; }

  operator double*() const		{ return base; }
#if 0
  // Some C++ compilers does not like this
  operator double**()			{ return &base; }
#else
  typedef double **dpp ;
  operator dpp() 			{ return &base; }
#endif

  void setsize(unsigned int n);
  void resize(unsigned int n);

  void print(const char* message) const;
  void print() const;

  double& operator[](unsigned int i) { return base[i];}
  friend GNU_CONST double norm_sqr(const Vect &v);
  friend GNU_CONST double norm1(const Vect &v);
  unsigned int capacity() const		{ return size; }

  friend /*static*/ GNU_CONST Vect operator+(const Vect &x, const Vect &y);
  friend /*static*/ GNU_CONST Vect operator-(const Vect &x, const Vect &y);
  friend /*static*/ GNU_CONST Vect operator*(const Vect &x, double a);
  friend /*static*/ GNU_CONST Vect operator*(double a, const Vect &x);
  friend /*static*/ GNU_CONST Vect operator/(const Vect &x, double a);

  inline Vect& operator=(const Vect &x);
  inline Vect& operator=(double x);
  inline Vect& operator+=(const Vect &x);
  inline Vect& operator-=(const Vect &x);
  inline Vect& operator*=(double a);
  inline Vect& operator/=(double a);
};

inline
Vect::Vect(unsigned int n)
{
  size = n;
  base = new double[n];
}

inline
Vect::Vect(const Vect& v)
{
  size = v.size;
  base = new double[size];
  for (unsigned i = 0; i < size; i++)
    base[i] = v.base[i];
}

inline GNU_CONST Vect
operator*(const Vect &w, double a)
  //#if defined(__GNUC__) && !defined(_G_NO_NRV)
  //     return v(w);
  //#endif       
{
  //#if !(defined(__GNUC__) && !defined(_G_NO_NRV))
  Vect v(w);
  //#endif
  for (unsigned i = 0; i < v.size; i++)
    v.base[i] *= a;
  return v;
}

inline GNU_CONST Vect
operator*(double a, const Vect &x)
{
  return x*a;
}

inline GNU_CONST Vect
operator/(const Vect &w, double a)
  //#if defined(__GNUC__) && !defined(_G_NO_NRV)
  //     return v(w);
  //#endif       
{
  //#if !(defined(__GNUC__) && !defined(_G_NO_NRV))
  Vect v(w);
  //#endif
  for (unsigned i = 0; i < v.size; i++)
    v.base[i] /= a;
  return v;
}

inline GNU_CONST Vect
operator+(const Vect &x, const Vect &y)
// #if defined(__GNUC__) && !defined(_G_NO_NRV)
//      return v(x);
// #endif
{
  //#if !(defined(__GNUC__) && !defined(_G_NO_NRV))
  Vect v(x);
  //#endif
  for (unsigned i = 0; i < v.size; i++) {
    v.base[i] += y.base[i];
  }
  return v;
}

inline GNU_CONST Vect
operator-(const Vect &x, const Vect &y)
// #if defined(__GNUC__) && !defined(_G_NO_NRV)
//      return v(x);
// #endif
{
  //#if !(defined(__GNUC__) && !defined(_G_NO_NRV))
  Vect v(x);
  //#endif
  for (unsigned i = 0; i < x.size; i++) {
    v.base[i] -= y.base[i];
  }
  return v;
}

inline Vect&
Vect::operator=(const Vect &x)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] = x.base[i];
  }
  return *this;
}

inline Vect&
Vect::operator=(double x)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] = x;
  }
  return *this;
}

inline Vect&
Vect::operator+=(const Vect &x)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] += x.base[i];
  }
  return *this;
}

inline Vect&
Vect::operator-=(const Vect &x)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] -= x.base[i];
  }
  return *this;
}

inline Vect&
Vect::operator*=(double a)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] *= a;
  }
  return *this;
}

inline Vect&
Vect::operator/=(double a)
{
  for (unsigned int i = 0; i < size; i++) {
    base[i] /= a;
  }
  return *this;
}

inline double
norm(const Vect &v)
{
  return sqrt(norm_sqr(v));
}

#endif
