/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: commands.cc,v 1.20 1994/11/01 12:52:38 bengt Exp bengt $ */

/*
 * This file defines decode_commands, which is doing the main job of decoding
 * commands in a script file (or stdin). It calls zzparse, defined through
 * cmdparse.y. It also defines miscellaneous support routines for the parser
 * and the scanner.
 */

#include "externals.h"
#include "cmdparse.h"
#include "init.h"

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#ifndef X_OK
#define X_OK    01
#endif

#define STR_EQUAL(s1, s2) ! strcmp((s1),(s2))
#define SCRIPT_LOG if (do_logging) va_log_event

/* Forward declarations */
static bool parse_and_close(bool initialize);
static bool find_store(UserParameter *param[], int size, const char *name,
		       bool stored);
void dump_symrecs(symrec *p);

bool lex_virgin = true;

/*
 * Holds the FILE scriptfile, NULL if we are currently not using one,
 * e.g. if we are taking commands from stdin.
 */
static FILE *scriptfile = NULL;
static bool scriptfile_is_pipe;

/* The symbol table: a chain of `struct symrec'.  */
static symrec *sym_table = NULL;

/* Table of mathematical functions for the symbol table */
static init arith_fncts[] = {
  // Believe all systems have these, don't test
  {"sin", sin},
  {"cos", cos},
  {"tan", tan},
  {"atan", atan},
  {"asin", asin},
  {"acos", acos},
  {"abs", fabs},
  {"fabs", fabs},
  {"ln", log},
  {"log", log},
  {"log10", log10},
  {"lg", log10},
  {"exp", exp},
  {"sqrt", sqrt},

#ifdef HAVE_RINT
  {"int", rint},
  {"rint", rint},
#endif

#ifdef HAVE_CEIL
  {"ceil", ceil},
#endif

#ifdef HAVE_FLOOR
  {"floor", floor},
#endif

#ifdef HAVE_EXP2
  {"exp2", exp2},
#endif

#ifdef HAVE_LOG2
  {"log2", log2},
#endif

#ifdef HAVE_EXP10
  {"exp10", exp10},
#endif

#ifdef HAVE_CBRT
  {"cbrt", cbrt},
#endif

#ifdef HAVE_ACOSH
  {"acosh", acosh},
#endif

#ifdef HAVE_ASINH
  {"asinh", asinh},
#endif

#ifdef HAVE_ATANH
  {"atanh", atanh},
#endif

#ifdef HAVE_ERF
  {"erf", erf},
#endif

#ifdef HAVE_ERFC
  {"erfc", erfc},
#endif

#ifdef HAVE_EXPM1
  {"expm1", expm1},
#endif

#ifdef HAVE_J0
  {"j0", j0},
#endif

#ifdef HAVE_J1
  {"j1", j1},
#endif

#ifdef HAVE_LGAMMA
  {"lgamma", lgamma},
#endif

#ifdef HAVE_LOG1P
  {"log1p", log1p},
#endif

#ifdef HAVE_Y0
  {"y0", y0},
#endif

#ifdef HAVE_Y1
  {"y1", y1},
#endif

  {"", NULL },
};

/****************************************************************/
#ifdef READLINE

// The following three functions are more or less copied verbatim form the
// readline manual.

// The readline.h is not prototyped, therefore declare needed funcs here.
extern "C" char *rl_readline_name;
extern "C" char *rl_line_buffer;
extern "C" int (*rl_attempted_completion_function)(char *, int, int);
extern "C" char **completion_matches(const char *text, 
				     char *(*entry_function) (const char *, int));

extern "C" char *xmalloc(size_t); // in readline

static char *
dupstr(const char *source) 
{
  char * copy = xmalloc(strlen(source) + 1);
  strcpy(copy, source);
  return copy;
}

/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char *
command_generator(const char *text, int state)
{
  extern char *script_commands[]; // Defined in lexical.ll

  static int list_index, len;
  char *name;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state) {
    list_index = 0;
    len = strlen(text);
  }

  /* Return the next name which partially matches from the command list. */
  while ((name = script_commands[list_index])) {
    list_index++;

    if (strncmp (name, text, len) == 0) {
      return (dupstr(name));
    }
  }

  /* If no names matched, then return NULL. */
  return(NULL);
}

/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
char **
gnans_completion(const char *text, int start, int /*end*/)
{
  char **matches;

  matches = NULL;

  /* Ignore whitespace at the start of the line */
  char *p = rl_line_buffer;
  while (*p == ' ' || *p == '\t') {
    start--;
    p++;
  }
  
  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  if (start == 0)
    matches = completion_matches(text, &command_generator);

  return(matches);
}

void
initialize_readline()
{
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "Gnans";

  /* Tell the completer that we want a crack first. */
  rl_attempted_completion_function =
    (int (*)(char *, int, int)) gnans_completion;

  // readline somehow turns off echo if the program runs in the background.
  // I turn it on manually here.
  // This code works, but I have no idea if it is "the right way" to do it.
  // Really, this is a bug in readline that should be fixed there.
  // Somehow, this doesn't seem to be needen on Solaris 2.

#if 0
#if !(defined(SYSV) || defined(SVR4))
  int success;
  struct sgttyb sg;
  success = ioctl(0, TIOCGETP, &sg);
  if (success == -1)
    gnans_error(impossible, "ioctl failed");
  sg.sg_flags |= ECHO;
  success = ioctl(0, TIOCSETP, &sg);
  if (success == -1)
    gnans_error(impossible, "ioctl failed");
#endif // ! SYSV (this should be fixed)

#if defined(sgi)
  int success;
  struct termios t;
  success = ioctl(0, TCGETS, &t);
  if (success == -1)
    gnans_error(impossible, "ioctl failed");
  t.c_lflag |= ECHO;
  t.c_cc[VEOF] = '\004';
  t.c_cc[VSUSP] = '\032';
  success = ioctl(0, TCSETS, &t);
  if (success == -1)
    gnans_error(impossible, "ioctl failed");
#endif
#endif
}

#endif				// READLINE

/* Support routines, first the ones for handling the symbol table */

/* Enters a new node as first element */
extern symrec *
putsym(const char *sym_name, int sym_type, bool ro)
{
  symrec *ptr = new symrec(sym_name, sym_type, ro, sym_table);
  sym_table = ptr;
  return ptr;
}

static symrec *
putsym(const char *sym_name, int sym_type, bool ro, double *d_ptr)
{
  symrec *p = putsym(sym_name, sym_type, ro);
  *p = d_ptr;
  return p;
}

/* Return a pointer to a symbol, given its name. */
extern symrec
*getsym(const char *sym_name)
{
  symrec *ptr;
  for (ptr = sym_table; ptr; ptr = ptr->next)
    if (STR_EQUAL(ptr->name, sym_name))
      return ptr;
  return NULL;
}

/* Free all memory allocated by the symbol table. */
static void
free_symbol_table()
{
  symrec *q, *p = sym_table;
#if 0
  if (debug_1)
    dump_symrecs(sym_table);
#endif
  while (p) {
    q = p;
    p = p->next;
    delete q;
  }
  sym_table = NULL;
}

/* Enter an array of UserParameter into the symbol table. */
static void
enter_gnans_array(UserParameter* param[], int size, bool readonly)
{
  int i;
  symrec *p;
  for (i = 0; i < size; i++)
    p = putsym(param[i]->name(), VAR, readonly, *param[i]);
}

/* Enter the states + their initial conditions into the symbol table */
static void 
enter_gnans_states()
{
  unsigned int i;
  symrec *p;
  char str[MAXSTRINGSIZE];
  for (i = 0; i < current_system.number_of_states; i++) {
    /* The state, readonly */
    p = putsym(current_system.states[i]->name(), VAR, true,
	       (double *) current_system.state_vector + i);
    //    p = (double *) current_system.state_vector + i;
    strcpy(str, current_system.states[i]->name());
    /* The initial condition, r/w */
    p = putsym(strcat(str, "@"), VAR, false, *current_system.states[i]);
  }
}

/* Enter the arithmetical functions (like sin) into the symbol table */
static void 
enter_gnans_arith_funcs()
{
  int i;
  symrec *ptr;
  for (i = 0; arith_fncts[i].fnct; i++) {
    ptr = putsym (arith_fncts[i].fname, FNCT, true);
    ptr->fnctptr = arith_fncts[i].fnct;
  }
}

/*
 * Initialize the symbol table by entering by entering everything than
 * should be there by initialization.
 */
static void 
init_table()
{
  enter_gnans_arith_funcs();
  enter_gnans_states();
  enter_gnans_array(internals, 1, true); /* System time, readonly */
  enter_gnans_array(internals + 1, number_of_internals - 1, false);
  enter_gnans_array(current_system.randoms,
		    current_system.number_of_randoms, false);
  enter_gnans_array(current_system.parameters,
		    current_system.number_of_parameters, false);
  enter_gnans_array(current_system.functions,
		    current_system.number_of_functions, true);
  putsym("ans", VAR, false);
}

extern void
update_symbol_table()
{
  free_symbol_table();
  init_table();
}

/* Print the prompt, provided that it makes sense. */
extern void 
do_prompt()
{
  if (!scriptfile 
#ifdef READLINE
      && ! (isatty(fileno(stdin)) && isatty(fileno(stdout)))
#endif
      )
    printf(cmd_prompt.c_str());
  fflush(stdout);
}

/*
 * Called by zzparse on parse error.
 * Aborts script execution but not interactive sessions.
 */
extern void 
zzerror(char *s)
{
  if (scriptfile)
    abort_script = true;
  gnans_error(parse, s);
}

/* Routines implementing commands */

/*
 * Sets the store status of a UserParameter. If append_trajectories is true,
 * we must write the trajectories to disk, otherwise they might be
 * inconsistent.
 */ 
extern void
store(const char *name, bool stored)
{
  if (! (
	 find_store(current_system.states,
		    current_system.number_of_states, name, stored)
	 || find_store(current_system.parameters,
		       current_system.number_of_parameters, name, stored)
	 || find_store(current_system.functions,
		       current_system.number_of_functions, name, stored)
	 || find_store(current_system.randoms,
		       current_system.number_of_randoms, name, stored)
	 || find_store(internals, number_of_internals, name, stored)))
    gnans_error(non_fatal, "Could not find variable ``%s''", name);
}

/*
 * Support routine for the routine above. Seaches an array of
 * UserParameters for the parameter. Returns true if found.
 */
static bool
find_store(UserParameter *param[], int size, const char *name, bool stored)
{
  int i;
  for (i = 0; i < size; i++)
    if (STR_EQUAL(param[i]->name(), name)) {
      storeset_dirty = storeset_dirty || param[i]->store_me() != stored;
      param[i]->store(stored);
      SCRIPT_LOG("%s %s", stored ? "Store" : "Unstore", name);
      if (append_trajectories)
	write_trajectories();
      return true;
    }
  return false;
}    

/* Toggles a global switch and invokes its triggering function. */
extern void
toggle_switch(const char *name, bool value)
{
  int i;
  for (i = 0; switches_table[i].ptr; i++) {
    if (strcasecmp(switches_table[i].name, name) == 0) {
      SCRIPT_LOG("toggle_switch (%s); %s -> %s",
		 switches_table[i].name,
		 *switches_table[i].ptr ? "true"  : "false",
		 value ? "true" : "false");
      *switches_table[i].ptr = value;
      if (switches_table[i].fcn)
	(*switches_table[i].fcn)();
      return;
    }
  }
  if (! switches_table[i].ptr)
    gnans_error(non_fatal, "Switch ``%s'' not found", name);
}

/****************************************************************/
/*
 * Decode the commands from a file with the prescribed name. 
 * If the name ends with a ``|'', this character is stripped, and the
 * remaning named is popen-ed. If the name is "", take commands from 
 * stdin instead.
 *
 * This is the main function.
 */
extern void 
decode_commands(const char *scriptfilename)
{
  int result;
  bool Xsaved = X;
  typedef struct stat struct_stat;
  struct_stat *buf = new struct_stat;

  char *realfilename = new char[strlen(scriptfilename) + 1];
  strcpy(realfilename, scriptfilename);
  if (scriptfilename[strlen(scriptfilename)-1] == '|') {
    scriptfile_is_pipe = true;
    realfilename[strlen(scriptfilename)-1] = '\0';
  } else
    scriptfile_is_pipe = false;

  if (! STR_EQUAL(scriptfilename, "") && ! STR_EQUAL(scriptfilename, "-")) {
    /* Non-empty filename */
    
    /* First find out if it is there */
    result = stat(realfilename, buf);
    if (result) {
      switch (errno) {
      case ENOENT:
      case ENOTDIR:
	gnans_error(non_fatal, "File ``%s'' does not exist", realfilename);
	return;
      case EACCES:
	gnans_error(non_fatal,
		    "You are not permitted to stat file ``%s''",
		    realfilename);
	return;
      case ELOOP:
      case ENAMETOOLONG:
	gnans_error(non_fatal, "Illegal file name ``%s''", realfilename);
	return;
      default:
	gnans_error(non_fatal, "Error occured trying to stat file ``%s''",
		    realfilename);
	return;
      }
    }

    if (S_ISDIR(buf->st_mode) || S_ISCHR(buf->st_mode)
	|| S_ISBLK(buf->st_mode)) {
      gnans_error(non_fatal, "File ``%s'' is not a regular file",
		  realfilename);
      return;
    }

    // If it ends with a ``|'', try to popen it
    if (scriptfilename[strlen(scriptfilename)-1] == '|') {
      // If relative file name, prepend ``./'', thus bypassing PATH
      if (realfilename[0] != '/' && realfilename[0] != '.') {
	char *s = new char[strlen(realfilename) + 3];
	strcpy(s, "./");
	strcpy(s + 2, realfilename);
	delete [] realfilename;
	realfilename = s;
      }
      scriptfile = popen(realfilename, "r");
      if (scriptfile) {
	SCRIPT_LOG("Opened ``%s'' for execution", realfilename);
      }
    } else {
      // Try to read it
      scriptfile = fopen(realfilename, "r");
      if (scriptfile)
	SCRIPT_LOG("Opened ``%s'' for reading", realfilename);
    }
    
    if (scriptfile)
      zzin = scriptfile;
    else {
      gnans_error(non_fatal, "Could not open script file");
      return;
    }
    if (X && workprocs)
      enable_buttons_busy();
  } else {
    /* Empty file name = stdin */
    SCRIPT_LOG("Taking commands from stdin");
    zzin = stdin;
    X = false;			/* Disable X while typing */
  }
  
  delete buf;
  delete [] realfilename;

  reset_cpu();  
  init_table();
  do_prompt();

  /* Flex (but not lex) needs to be restarted */
  if (lex_virgin)
    lex_virgin = false;
  else
    zzrestart(zzin);

  if (X && workprocs)
    register_work_proc(false, &parse_and_close);
  else
    parse_and_close(true);
  X = Xsaved;
}

/*
 * The bulk of the work is done here. This is called (in general many
 * times for one execution) thru the workproc.
 * Its argument states if it is called for the first time or not.
 * It returns either if the task is finished (returning cleaning up
 * and returning  true), if an error occured (returning cleaning up
 * and returning  true), or if it was interrupted by the clock, in which
 * case it returns false and expects to be called again.
 */
static bool
parse_and_close(bool initialize)
{
  int result;
  if (initialize) {
    running_script = !(zzin == stdin && isatty(fileno(stdin))
		       && isatty(fileno(stdout)));
#ifdef READLINE
    if (!running_script)
      initialize_readline();
#endif // READLINE
    abort_script = false;
  }
  if (running_integration)
    if (! continue_integration(false))
      return false;
  result = zzparse();
  switch (result) {
  case 0:
    /* Finished */
    log_event("Script ended successfully");
    if (!scriptfile && isatty(fileno(stdin)) && isatty(fileno(stdout))) {
      /* User pressed ^D, output a newline for prettyness */
      printf("\n");
      fflush(stdout);
    }
    break;
  case 1:
    /* Nonrecoverable error */
    log_event("Script ended with error or was aborted");
    break;
  case 2:
    /* Clock interupt */
    return false;
  }
  
  running_script = false;
  if (scriptfile)
    if (scriptfile_is_pipe)
      pclose(scriptfile);
    else
      fclose(scriptfile);
  scriptfile = NULL;
  free_symbol_table();
  if (X && ! workprocs) {
    enable_buttons_idle();
    indicate_progress();
  }
  return true;
}
/*****************************************************************/
/* Debugging stuff */
void
symrec::print()
{
  gnans_error(debug, "%-20s %g\t%s\t%d",
	      name, *var, readonly ? "ro" : "r/w", type);
}

void
dump_symrecs(symrec *p)
{
  if (!p)
    return;
  p->print();
  dump_symrecs(p->next);
}
  
/****************************************************************/
symrec::symrec(const char *sym_name, int sym_type, bool ro,
	       symrec *next_ptr)
{
  precious_var = false;
  name = new char[strlen(sym_name) + 1];
  var = new double;
  *var = gnans_NaN();
  strcpy(name, sym_name);
  type = sym_type;
  readonly = ro;
  next = next_ptr;
}

symrec::~symrec()
{
  delete [] name;
  if (!precious_var) delete var;
}

symrec&
symrec::operator=(double *v_ptr)
{
  if (!precious_var)
    delete var;
  var = v_ptr;
  precious_var = true;
  return *this;
 }
