/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: display.cc,v 1.16 1994/08/19 18:38:31 bengt Exp $ */

/* Here we collect functions displaying or dumping data. */

#define USETEXTWIDGET		/* Textwidget or labelwidget for display? */

#include "externals.h"
#include "CIntegrator.h"
#include "Resource.h"
#include "cmdparse.h"
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#ifdef USEXAW3DINCLUDES

#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Cardinals.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Dialog.h>
#include <X11/Xaw3d/Toggle.h>
#include <X11/Xaw3d/AsciiText.h>

#else

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>

#endif

/* How are the displays to be formatted? */
#define PARAMETER_FMT "%20s = %8.5g\t\t\t%s\n"
#define RESOURCE_FMT  "%20s = %8.5g\n"
#define INTEGER_FMT   "%20s = %8d\n"
#define STRING_FMT    "%20s = %s\n"
#define STATE_FMT     "%20s = %8.5g\t(Initial: %g)\t%s\n"

#define STR_EQUAL(a,b) ( ! strcmp((a),(b)))

/* Global (for the file) variable holding the string under construction */
static char *display_string;

/* Pointer to the current position in display_string */
static char *string_ptr;

/* This function appends the string given as argument to the display_string */
static void
append(char *str)
{
  do {
    *string_ptr++ = *str++;
  } while (*str);
  *string_ptr = '\0';
}

inline static char *
stored_str(bool stored)
{
  if (stored)
    return "(Stored)";
  else
    return "";
}

inline static void
display_string_value(char *name, const char *value)
{
  char str[MAXSTRINGSIZE];
  sprintf(str, STRING_FMT, name, value);
  append(str);
}

inline static void
display_string_value(char *name, std::string value)
{
  display_string_value(name, value.c_str());
}

/*
 * Displays an entire array of user_parameters.
 */
static void 
display_array(UserParameter *arr[], int size)
{
  if (arr) {
    int i;
    char str[MAXSTRINGSIZE];
    for (i = 0; i < size; i++) {
      sprintf(str, PARAMETER_FMT, arr[i]->name(), double(*arr[i]),
	      stored_str(arr[i]->store_me()));
      append(str);
    }
  }
}

/*
 * Displays the resource (cpu, memory) usage.
 */
static void
display_resources()
{
  char str[MAXSTRINGSIZE];
  Resource rusage;
  sprintf(str, RESOURCE_FMT, "User Time", rusage.user_seconds());
  append(str);
  sprintf(str, RESOURCE_FMT, "System Time", rusage.system_seconds());
  append(str);
#ifdef WORKING_MEMORY_GAUGE
  sprintf(str, INTEGER_FMT, "KBytes", rusage.kilobytes());
  append(str);
#endif
}

/*
 * Displays all internals. Take some care not to display irrelevant
 * information, such as used integration algorithm by discrete time systems.
 * The argument states how many internals should be displayed. (Yes, bad
 * design. Watch out when updating the program.)
 */
static void 
display_internals(unsigned int number)
{
  unsigned int i;
  char str[MAXSTRINGSIZE];
  for (i = 0; i < number; i++) {
    if ((current_system.system_type == discrete
	 && (STR_EQUAL(internals[i]->name(), "timestep")
	     || STR_EQUAL(internals[i]->name(), "wienertick")
	     || STR_EQUAL(internals[i]->name(), "relerror")
	     || STR_EQUAL(internals[i]->name(), "abserror")))
	|| (current_system.number_of_wieners == 0
	    && STR_EQUAL(internals[i]->name(), "wienertick"))
	|| (STR_EQUAL(internals[i]->name(), "methodflag")
	    && !(expert_mode || development))
	|| (STR_EQUAL(internals[i]->name(), "algorithm_parameter_1")
	    && !development)
	|| (STR_EQUAL(internals[i]->name(), "algorithm_parameter_2")
	    && !development)
	|| (STR_EQUAL(internals[i]->name(), "algorithm_parameter_3")
	    && !development))
      /* Don't print it */ ;
    else { 
      sprintf(str, PARAMETER_FMT,
	      internals[i]->name(),
	      double(*internals[i]),
	      stored_str(internals[i]->store_me()));
      append(str);
    }
  }
  if (number == number_of_internals
      && current_system.system_type != discrete) {
    sprintf(str, STRING_FMT, "algorithm", current_algorithm->name());
    append(str);
  }
}

/*
 * Displays all states.
 */
static void 
display_state_array(double states[], UserParameter *arr[],
				int size)
{
  int i;
  char str[MAXSTRINGSIZE];
  for (i = 0; i < size; i++) {
    sprintf(str, STATE_FMT,
	    arr[i]->name(), states[i], double(*arr[i]),
	    stored_str(arr[i]->store_me()));
    append(str);
  }
}

// Builds up a string containing the text of the display.
// If arg is "internals", resources", "states", "parameters", "functions",
// or "times", display these.  Else dump it all.

// Note: the caller must delete [] the return value.
extern char *
create_display_string(int what_to_display)
{
  bool do_functions = ! what_to_display || what_to_display == FUNCTIONS;
  bool do_internals = ! what_to_display || what_to_display == INTERNALS;
  bool do_resources = ! what_to_display || what_to_display == RESOURCES;
  bool do_parameters = ! what_to_display || what_to_display == PARAMETERS;
  bool do_randoms = ! what_to_display || what_to_display == RANDOMS;
  bool do_states = ! what_to_display || what_to_display == STATES;
  bool do_times = what_to_display == TIMES;

  display_string = new char[(5
			     + number_of_internals
			     + current_system.number_of_states
			     + current_system.number_of_parameters
			     + current_system.number_of_functions
			     + current_system.number_of_randoms)
			    * MAXSTRINGSIZE];
  string_ptr = display_string;
  
  if (! what_to_display) {
    display_string_value("Gnans Version", version_string);
    display_string_value("System Name", current_system.system_name);
  }

  if (do_times) 
    display_internals(number_of_internals - 6);
  /* Watch out when updating the program!! */

  if (do_internals)
    display_internals(number_of_internals);
  if (do_states)
    display_state_array(current_system.state_vector,
			current_system.states,
			current_system.number_of_states);
  if (do_parameters)
    display_array(current_system.parameters,
		  current_system.number_of_parameters);
  if (do_functions)
    display_array(current_system.functions,
		  current_system.number_of_functions);
  if (do_randoms)
    display_array(current_system.randoms,
		  current_system.number_of_randoms);
  if (do_resources)
    display_resources();

  return display_string;
}

/*
 * The main function.
 * 
 * Display everything "interesting", on tty or in a popup.
 * If arg is "internals", resources", "states", "parameters", "functions",
 * or "times", display these.  Else dump it all.
 */
extern void 
Display_vars(int what_to_display)
{
  char title[MAXSTRINGSIZE];
  char *display_string = create_display_string(what_to_display);
  va_log_event("Display_vars");
  if (X) {
    Widget popup, dismiss, box;
    popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
				 toplevel,
				 XtNtitle, make_popup_title(title, "Display"),
				 NULL);
    box = XtVaCreateManagedWidget("box", boxWidgetClass, popup, NULL);
#ifdef USETEXTWIDGET
    XtVaCreateManagedWidget("display", asciiTextWidgetClass, box,
				   XtNstring, (char *) display_string,
				   NULL);
#else
    XtVaCreateManagedWidget("label", labelWidgetClass, box,
				   XtNlabel, (char *) display_string, NULL);
#endif
    dismiss = XtVaCreateManagedWidget("dismiss", commandWidgetClass,
				      box, NULL);
    XtAddCallback(dismiss, XtNcallback, destroy_inquiry_popup,
		  (XtPointer) &popup);
    popup_position(popup, toplevel, XtGrabNone, false);
  } else {
    fprintf(stdout, "%s", display_string);
  }
  delete [] display_string;
}

// Called during crash, to save as much as possible.
extern void 
emergency_dump()
{
  fprintf(stderr, "Emergency dump\n");
  if (! do_logging)
    open_log();
  log_event("****************EMERGENCY DUMP****************");
  display_in_logfile();
  close_log();
  unlink(GNANSLOCKFILENAME);
}
