/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: formats.cc,v 1.16 1994/08/19 18:38:31 bengt Exp $ */

/*
 * This function contains everything that is specific to the different store
 * formats. If adding a new format, only this file needs to be changed.
 */

#include "externals.h"
#include "utillib.h"
#include <fstream>
#include <stdlib.h>
#include <locale.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

/* Forward declarations */
static bool trueism(bool complain);
static bool obj_check(bool complain);
static bool float_check(bool complain);
static bool float_ni_check(bool complain);

static bool ascii_write();
static bool ascii_ni_write();
static bool csv_write();
static bool obj_write();
static bool raw_double_write();
static bool raw_double_ni_write();
static bool raw_float_write();
static bool raw_float_ni_write();
static bool matlab_write();
static bool maple_write();
static bool null_write();

#define MAPLE_V_2

/*
 * (Global) Table of store formats. The type StoreFormat is defined
 * (and commented) in gnans.h.
 */

StoreFormat store_format_table[] = {
/* 0 */
  {NULL, "Matlab", "mat", &trueism, &matlab_write,
     "Store in a file which can be directly read by Matlab\n"
     "(preserving variable names (after truncating to 19 characters))."},

/* 1 */
{NULL, "Maple", "maple", &trueism, &maple_write,
     "Store in a file which can be directly read by Maple\n"
"(preserving variable names)."
#ifdef MAPLE_V_2
"\nRequires Maple V Release 2."
#endif
},

/* 2 */
{NULL, "Ascii", "ascii", &trueism, &ascii_write,
   "Store in an ascii text-file, one row per data point."},

/* 3 */
{NULL, "Ascii (noninterlaced)", "ascii", &trueism, &ascii_ni_write,
   "Store in an ascii text-file.\n"
"The variables are stored in non-interleaved format,\n"
"i.e. after one another."},

/* 4 */
{NULL, "Raw/double", "bin", &trueism, &raw_double_write,
   "Store in a raw (machine dependent) file as doubles."},

/* 5 */
{NULL, "Raw/double (noninterlaced)", "bin",
   &trueism, &raw_double_ni_write,
   "Store in a raw (machine dependent) file as doubles.\n"
"The variables are stored in non-interleaved format,\n"
"i.e. after one another."},

/* 6 */
{NULL, "Raw/float", "bin", &float_check, &raw_float_write,
   "Store in a raw (machine dependent) file as floats."},

/* 7 */
{NULL, "Raw/float (noninterlaced)", "bin",
   &float_ni_check, &raw_float_ni_write,
   "Store in a raw (machine dependent) file as floats.\n"
"The variables are stored in non-interleaved format,\n"
"i.e. after one another."},

/* 8 */
{NULL, "Obj", "obj", &obj_check, &obj_write,
   "Object data format for 3D objects"},

/* 9 */
{NULL, "CSV", "csv", &trueism, &csv_write,
   "Store in an ascii text-file, one row per data point,\n"
   "with a semicolon `;' between the numbers.\n"
   "This can be used, e.g., by office programs."},

/* 10 */
{NULL, "Null", "", &trueism, &null_write, "Don't store anything at all"},

{NULL, NULL, NULL, NULL, NULL, NULL}
};

/****************************************************************/
// Miscellaneous checking functions

#define COMPLAIN(s) if (complain) gnans_error(non_fatal, s);

#if 0
static bool
not_implemented(bool complain)
{
  COMPLAIN("Sorry, the selected format is not yet implemented");
  return false;
}
#endif

static bool
trueism(bool /*complain*/)
{
  return true;
}

static bool 
obj_check(bool complain)
{
  if (number_of_stores != 3) {
    COMPLAIN("The selected format requires exactly three"
	     " variables to be stored");
    return false;
  }
  return true;
}

static bool
float_check(bool /*complain*/)
{
  if (sizeof(float) != sizeof(int))
    gnans_error(impossible, "Internal error");
  return true;
}

static bool
float_ni_check(bool /*complain*/)
{
  if (sizeof(float) != sizeof(int))
    gnans_error(impossible, "Internal error");
  return true;
}

//////////////// Write info files
static bool
do_wiw_file()
{
  std::ofstream wiwfile(GNANSWIWFILENAME/*, ios::out*/);
  if (!wiwfile || wiwfile.bad()) {
    gnans_error(non_fatal, "Cannot open file %s\n", GNANSWIWFILENAME);
    return false;
  }

  // Write the wiw-file
  wiwfile << "# Generated by Gnans version " << (char *) version_string << "\n";
  wiwfile << "# at " << getdate() << " by " << getlogin() << "@"
    << gethostname_as_string() << "\n";
  wiwfile << "# Variables: \n";
  
  unsigned int i;
  for (i = 0; i < number_of_internals; i++) // t and so on...
    if (internals[i]->store_me()) {
      wiwfile.width(10); /*wiwfile.setf(ios::left);*/
      wiwfile << internals[i]->name();
    }
  for (i = 0; i < current_system.number_of_states; i++)
    if (current_system.states[i]->store_me()) {
      wiwfile.width(10); // wiwfile.flags(ios::left);
      wiwfile << current_system.states[i]->name();
    }
  for (i = 0; i < current_system.number_of_functions; i++) 
    if (current_system.functions[i]->store_me()) {
      wiwfile.width(10); //wiwfile.flags(ios::left);
      wiwfile << current_system.functions[i]->name();
    }
  for (i = 0; i < current_system.number_of_parameters; i++) 
    if (current_system.parameters[i]->store_me()) {
      wiwfile.width(10); //wiwfile.flags(ios::left);
      wiwfile << current_system.parameters[i]->name();
    }
  for (i = 0; i < current_system.number_of_randoms; i++) 
    if (current_system.randoms[i]->store_me()) {
      wiwfile.width(10); //wiwfile.flags(ios::left);
      wiwfile << current_system.randoms[i]->name();
    }
  
  wiwfile << "\n";		// Closed by destructor
  return true;
}

static bool
do_parameter_file()
{
  if (! store_parameters)
    return true;
  std::ofstream parameterfile(GNANSPARAMETERFILENAME);
  if (!parameterfile || parameterfile.bad()) {
    gnans_error(non_fatal, "Cannot open file %s\n", GNANSPARAMETERFILENAME);
    return false;
  }

  // Write the parameter-file
  parameterfile << "# Generated by Gnans version " << (char *) version_string << "\n";
  parameterfile << "# at " << getdate() << " by " << getlogin() << "@"
    << gethostname_as_string() << "\n";
  parameterfile << "# Parameters:\n";
    
  for (unsigned i = 0; i < current_system.number_of_parameters; i++) {
    parameterfile.width(10); //parameterfile.flags(ios::left);
    parameterfile << current_system.parameters[i]->name()
      << " = " << (double) *current_system.parameters[i] // should be a method <<
	<< ";\n";
  }
  return true;
}

/****************************************************************/
/*
 * Format specific functions for writing
 */

#define BARF_ON_FAILURE if (result == EOF) return false
#define CHECK(a, b) if ((a) != (b)) return false
#define EXECUTE_CHECK(cmd) if (! cmd()) return false;

static bool
ascii_write()
{
  unsigned int i,j;
  int result;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (i = 0; i < sample_no; i++) {
    for (j = 0; j < number_of_stores; j++) {
      result = fprintf(stores[0].file, "%-10g ", stores[j].trajectory[i]);
      BARF_ON_FAILURE;
    }
    result = fprintf(stores[0].file,"\n");
    BARF_ON_FAILURE;
  }
  return true;
}

static bool
ascii_ni_write()
{
  unsigned int i, j;
  int result;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (j = 0; j < number_of_stores; j++)
    for (i = 0; i < sample_no; i++) {
      result = fprintf(stores[0].file, "%-10g\n", stores[j].trajectory[i]);
      BARF_ON_FAILURE;
    }
  return true;
}

static bool
csv_write()
{
  unsigned int i,j;
  int result;
  char old_lc_numeric[100];
  setlocale(LC_NUMERIC, getenv("LC_NUMERIC"));
  //EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);

  fprintf(stores[0].file, "# Generated by Gnans version %s\n",
	  (char *) version_string);
  fprintf(stores[0].file, "# at %s by %s@%s\n", getdate(), getlogin(),
	  gethostname_as_string().c_str());
  fprintf(stores[0].file, "# Variables: \n# ");
  for (i = 0; i < number_of_internals; i++) // t and so on...
    if (internals[i]->store_me()) {
      fprintf(stores[0].file, "%s;", internals[i]->name());
    }
  for (i = 0; i < current_system.number_of_states; i++)
    if (current_system.states[i]->store_me()) {
      fprintf(stores[0].file, "%s;", current_system.states[i]->name());
    }
  for (i = 0; i < current_system.number_of_functions; i++) 
    if (current_system.functions[i]->store_me()) {
      fprintf(stores[0].file, "%s;", current_system.functions[i]->name());
    }
  for (i = 0; i < current_system.number_of_parameters; i++) 
    if (current_system.parameters[i]->store_me()) {
      fprintf(stores[0].file, "%s;", current_system.parameters[i]->name());
    }
  for (i = 0; i < current_system.number_of_randoms; i++) 
    if (current_system.randoms[i]->store_me()) {
      fprintf(stores[0].file, "%s;", current_system.randoms[i]->name());
    }
  fprintf(stores[0].file, "\n");

  for (i = 0; i < sample_no; i++) {
    for (j = 0; j < number_of_stores; j++) {
      result = fprintf(stores[0].file,
		       j < number_of_stores - 1 ? "%g;" : "%g\n",
		       stores[j].trajectory[i]);
      BARF_ON_FAILURE;
    }
    BARF_ON_FAILURE;
  }
  return true;
}

#define OBJ_SCALING 100
static bool
obj_write()
{
  unsigned int i;
  int result;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  fprintf(stores[0].file, "%d\t%d\n", sample_no, sample_no - 1);
  for (i = 0; i < sample_no; i++) {
    result = fprintf(stores[0].file, "%g %g %g\n",
		     (OBJ_SCALING*stores[0].trajectory[i]),
		     (OBJ_SCALING*stores[1].trajectory[i]),
		     (OBJ_SCALING*stores[2].trajectory[i]));
    BARF_ON_FAILURE;
  }
  for (i = 0; i < sample_no - 1; i++) {
    result = fprintf(stores[0].file, "%d %d\n", i, i + 1);
    BARF_ON_FAILURE;
  }
  return true;
}

static bool
raw_double_write()
{
  unsigned int i, j;
  int result;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (i = 0; i < sample_no; i++)
    for (j = 0; j < number_of_stores; j++) {
      result = fwrite((char *) &(stores[j].trajectory[i]),
		      sizeof(double), 1,
		      stores[0].file);
      CHECK(result, 1);
    }
  return true;  
}

static bool
raw_double_ni_write()
{
  unsigned int j;
  size_t result;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (j = 0; j < number_of_stores; j++) {
    result = fwrite((char *) stores[j].trajectory, sizeof(double), sample_no,
		    stores[0].file);
    CHECK(result, sample_no);
  }
  return true;
}

static bool
raw_float_write()
{
  unsigned int i, j;
  int result;
  float dummy;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (i = 0; i < sample_no; i++)
    for (j = 0; j < number_of_stores; j++) {
      dummy = (float) stores[j].trajectory[i];
      result = fwrite((char *) &dummy, sizeof(float), 1, stores[0].file);
      CHECK(result, 1);
    }
  return true;
}

static bool
raw_float_ni_write()
{
  unsigned int i, j;
  int result;
  float dummy;
  EXECUTE_CHECK(do_wiw_file);
  EXECUTE_CHECK(do_parameter_file);
  for (j = 0; j < number_of_stores; j++) {
    for (i = 0; i < sample_no; i++) {
      dummy = (float) stores[j].trajectory[i];
      result = fwrite((char *) &dummy, sizeof(float), 1, stores[0].file);
      CHECK(result, 1);
    }
  }
  return true;
}

static bool
null_write()
{
  return true;
}

// Reference: Matlab External References Guide, p 29f
// This may not work on Vaxes or Crays
#ifndef MATLABFILETYPE
#ifdef WORDS_BIGENDIAN
#define MATLABFILETYPE 1000
#else
#define MATLABFILETYPE 0000
#endif
#endif

static int
min(int a, int b)
{
  return a < b ? a : b;
}

// Identifiers in Matlab should be at most this long, counting trailing 0.
//  Grrr...
const int mxMAXNAM = 20;

typedef struct {
  long type;			/* type */
  long mrows;			/* row dimension */
  long ncols;			/* column dimension */
  long imagf;			/* flag indicating imag part */
  long namlen;			/* name length (including NULL) */
} Fmatrix;

static bool
matlab_write()
{
  unsigned int j;
  int result;
  Fmatrix x;

  x.type = MATLABFILETYPE;
  x.mrows = (int) sample_no;
  x.ncols = 1;
  x.imagf = false;
	
  for (j = 0; j < number_of_stores; j++) {
    x.namlen = min(stores[j].name.length() + 1, mxMAXNAM);

    result = fwrite(&x, sizeof(Fmatrix), 1, stores[0].file);
    BARF_ON_FAILURE;
    result = fwrite(stores[j].name.c_str(), sizeof(char),
		    (int) x.namlen - 1, stores[0].file);
    BARF_ON_FAILURE;
    result = fwrite("", sizeof(char), 1, stores[0].file);
    BARF_ON_FAILURE;
    result = fwrite(stores[j].trajectory, sizeof(double),
		    (int) sample_no, stores[0].file);
    BARF_ON_FAILURE;
  }

  if (store_parameters) {
    x.mrows = 1;
    for (j = 0; j < current_system.number_of_parameters; j++) {
      x.namlen = min(strlen(current_system.parameters[j]->name()) + 1,
		     mxMAXNAM);
      
      result = fwrite(&x, sizeof(Fmatrix), 1, stores[0].file);
      BARF_ON_FAILURE;
      result = fwrite((const char *) current_system.parameters[j]->name(),
		      sizeof(char), (int) x.namlen - 1, stores[0].file);
      BARF_ON_FAILURE;
      result = fwrite("", sizeof(char), 1, stores[0].file);
      BARF_ON_FAILURE;
      result = fwrite((double*)(*current_system.parameters[j]),
		      sizeof(double), 1, stores[0].file);
      BARF_ON_FAILURE;
    }
  }
  return true;
}

static bool
maple_write()
{
  unsigned int i, j;
  int result;
  for (j = 0; j < number_of_stores; j++) {
    result = fprintf(stores[0].file,"%s := array(1..%d, [\n",
		     stores[j].name.c_str(), (int) sample_no);
    BARF_ON_FAILURE;
    for (i = 0; i < sample_no - 1; i++) {
#ifdef MAPLE_V_2
      result = fprintf(stores[0].file,"%g,\n", stores[j].trajectory[i]);
#else
      if (abs(stores[j].trajectory[i]) < 10000
	  && abs(stores[j].trajectory[i]) < 0.0001)
	result = fprintf(stores[0].file,"%f,\n", stores[j].trajectory[i]);
      else
	// Write it in the Float(mantissa, exponent) format, to be implemented
	result = fprintf(stores[0].file,"%f,\n", stores[j].trajectory[i]);
#endif
      BARF_ON_FAILURE;
    }
    result = fprintf(stores[0].file,"%f]):\n",
		     stores[j].trajectory[sample_no - 1]);
    BARF_ON_FAILURE;
  }

  if (store_parameters)
    for (j = 0; j < current_system.number_of_parameters; j++) {
      result = fprintf(stores[0].file, "%s=%g:\n",
		       (const char *) current_system.parameters[j]->name(),
		       (double)(*current_system.parameters[j]));
      BARF_ON_FAILURE;
    }

  return true;
}
