/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: globals.cc,v 1.20 1994/08/19 18:38:31 bengt Exp $

//
// This file contains all global variables, with some exceptions listed at
// the end of the file.


#include "externals.h"
#include "DIntegrator.h"
#ifdef RICCATI
#include "RIntegrator.h"
#endif

// The name of the program
std::string progname;

// The toplevel Widget
Widget toplevel = 0;

// X or tty interface?
bool X = false;

// Logging commands?
bool do_logging = true;

// Should also commands executed from scripts be logged?
bool log_script_commands = false;

// Should new runs be appended to the old trajectories, 
// rather than replace them?
bool append_trajectories = false;

// Should the first value be store?
bool store_first_value = true;

// Should states and functions be stored per default?
bool store_as_default = true;

// Should parameters be written to file at the end of a run?
bool store_parameters = false;

// Should the user be treated like expert
// (allowed to do thing only experts should try)?
bool expert_mode = false;

// The following two flags are used in two ways: if read, tells whether an
// integration (a script) is in progress. If assigned false, the integration
// (the script) should be halted as soon as possible.
bool running_integration = false;

// This indicates if a script is being run, however not from a terminal
bool running_script = false;

// Setting this true indicates that a running script should be aborted
bool abort_script;

// If true, a floating point error occured (or a terminate call in gnanslib)
// Then don't store last point, and prohibit continuation
bool floating_point_error_occured = false;

// Holds a debug code, see debug.cc
int gnansdebug = 0;

/**************** Debugging flags *****************/

// Note: These are for debugging and development. The absolutely ``correct''
// implementation is not guaranteed. If they don't do what you want, it is 
// a feature, not a bug...


// Trace workprocs and alarms
bool trace_workproc = false;

// Trace all calls to/from integration routines
bool trace_calls = false;

// Trace popups etc
bool trace_widgets = false;

// Trace all calls to malloc and free?
bool malloc_debug = false;

// These can be used during debugging
bool debug_1 = false;
bool debug_2 = false;
bool debug_3 = false;

// This can be used while testing integration algorithms
bool trace_algorithms = false;

/***************************************************************/

// Should workprocs be used, instead of hanging the program during 
// integration/script execution? (Normallu true, basically for debugging)
bool workprocs = true;

//
// Setting this true (via command line options or resource) gives a few
// more buttons for debugging etc (like a button for core dump)
bool development = false;

// The SIGALRM handler sets this true, indicating that ongoing integration
// should be halted as soon as possible.
bool has_been_alarmed = false;

// store format to use (Index for current format in table)
unsigned int store_format = 0;		// Should be more customizable

unsigned int number_of_stores = 0;

// All parameters etc associated with the currently loaded system.
GnansSystem current_system;

// The index of the current trajectory point to be saved
unsigned int sample_no = 0;

// How many points have I, up until now, malloced for the trajectories?
unsigned int trajectories_current_size = 0;

// The current algorithm
Integrator *current_algorithm = NULL;

// Default script file (used as default in dialogs)
std::string scriptfilename;

// File name corresponding to the currently loaded (or to be loaded) system
std::string system_filename;

// Has the user fiddled with the store status of the user parameters since
// last integration?
bool storeset_dirty = false;

FILE *Store::file;

//***************************************************************/
// The internal parameters
UserParameter systemtime(NULL, "The system time.", true);
UserParameter starttime("starttime", "The start time for a run.");
UserParameter stoptime("stoptime", "The ending time for a run.");
UserParameter timestep("timestep",
		       "For algorithms with fixed step size, the step size.\n"
		       "For algorithms with automatic step size selection, "
		       "the communication interval.");
UserParameter storeincrement("storeincrement",
			      "Lower bound for the time between stored "
			      "data points.");
UserParameter wienertick("wienertick",
			  "Discretization step for Wiener processes.");
UserParameter relerror("relerror",
			"The relative error. "
			"Exact meaning is algorithm-dependent.");
UserParameter abserror("abserror",
			"The absolute error. "
			"Exact meaning is algorithm-dependent.");
UserParameter method_flag("methodflag",
			   "Method flag for the integrator. "
			   "Exact meaning is algorithm-dependent.");
UserParameter algorithm_parameter_1("algorithm_parameter_1",
				     "Parameter for experimenting with the integrator.");
UserParameter algorithm_parameter_2("algorithm_parameter_2",
				     "Parameter for experimenting with the integrator.");
UserParameter algorithm_parameter_3("algorithm_parameter_3",
				     "Parameter for experimenting with the integrator.");

/****************************************************************/
// Arrays of all internal Gnans parameters
UserParameter *internals[number_of_internals+1] = 
{
  &systemtime,
  &starttime,
  &stoptime,
  &timestep,
  &storeincrement,
  &wienertick,
  &relerror,
  &abserror,
  &method_flag,    
  &algorithm_parameter_1,    
  &algorithm_parameter_2,    
  &algorithm_parameter_3,
  NULL
  };

/****************************************************************/

// An array of quantities actually to be stored
Store *stores = NULL;

// This determines how floating point numbers are printed
std::string floating_point_format;

// Default store format
std::string default_store_format;

// The prompt presented to the user during terminal sessions
std::string cmd_prompt;

// What algorithm should be selected before the user has selected one?
std::string default_algorithm;

// ... if the system is a stochastic system (= has Wiener processes)
std::string default_stochastic_algorithm;

// The default time step to be used in the algorithm above
double default_stochastic_timestep = 0;

// An array of the switches which should be presented to the user.
// If the user has selected -development on the command line
// (or flipped the development switch), (s)he will get 
// switches_table_development, otherwise switches_table_standard.
Switch *switches_table;

#define STANDARD_SWITCHES  \
{NULL, &do_logging, "Logging", &update_log_status, 			\
   "If true, all user invoked actions will be logged."},                \
{NULL, &log_script_commands, "Log Script Commands", NULL,               \
   "If true and logging is enabled, also script commands are logged."}, \
{NULL, &append_trajectories, "Append Trajectories", &append_trajectories_hook,\
   "If true, point computed at new runs are appended to the trajectories,"\
   "rather than replacing them."},                                      \
{NULL, &store_first_value, "Store First Value", NULL,			\
   "If true, the first value of the trajectory are always stored."      \
   "If false, the first stored point is at least storeincrement time units"\
   "larger than starttime (less than, if integrating backwards)."},     \
{NULL, &store_as_default, "Store As Default", NULL,			\
   "If true, all states and functions are marked stored when a system is loaded."},\
{NULL, &store_parameters, "Store Parameters", NULL,                   \
   "If true, parameters in systems are stored in files at the end of a run. " \
   "Exactly how depends on the store format selected."},\
{NULL, &expert_mode, "Expert Mode", NULL,          			\
   "If true, allows the user to do a few things only experienced"       \
   "user should attempt."},                                             \
{NULL, &development, "Development", &update_development_status,		\
   "Enable a few switches and buttons only for interest by debugging"   \
   "and development."}

Switch switches_table_standard[] = {
  STANDARD_SWITCHES,                                 
    {NULL, NULL, NULL, NULL, NULL}
};

Switch switches_table_development[] = {
  STANDARD_SWITCHES,
  {NULL, &workprocs, "workprocs", &update_workproc_status,
     "Use workprocs instead of foreground processing"},
  {NULL, (bool * ) &zz_flex_debug, "trace scanner", NULL,
     "Debugging the scanner"},
  {NULL, (bool * ) &zzdebug, "trace parser", NULL, "Debugging the parser"},
  {NULL, &malloc_debug, "trace malloc", NULL, "Tracing mallocs"},
  {NULL, &trace_workproc, "trace workproc", NULL, "Tracing workproc"},
  {NULL, &trace_calls, "trace calls", NULL, "Tracing calls"},
  {NULL, &trace_widgets, "trace widgets", NULL, "Tracing widget creation etc"},
  {NULL, &trace_algorithms, "trace algorithms", NULL,
   "Tracing of some interation algoriths"},

  {NULL, &debug_1, "Debug 1", NULL, ""},
  {NULL, &debug_2, "Debug 2", NULL, ""},
  {NULL, &debug_3, "Debug 3", NULL, ""},

  {NULL, NULL, NULL, NULL, NULL}
};

/****************************************************************/
// algorithms[] is an array, containing the integration routines
ContinuousIntegrator **algorithms;
DiscreteIntegrator discrete_time_integrate;
#ifdef RICCATI
RiccatiIntegrator riccati_integrate;
#endif
/****************************************************************/
// There is also a global store_format_type store_format_table[] in formats.cc
// version.cc contains version_string and compiled_string
