/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: log.cc,v 1.6 1994/08/24 09:58:29 bengt Exp $ */

/*
 * This file contains logging functions. The most useful function is
 * va_log_event, which takes a variable number of arguments, like printf
 *
 * Note: Callback functions which calls other functions for performing
 * the main task (e.g. Start(), which calls integrate()) should not do 
 * the logging themselves, but let the called function do the logging
 */

#include "externals.h"
#include "utillib.h"
#include <stdarg.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

extern "C" char * cuserid ( char *string ); // Flabeg

#define STR_EQUAL(s1, s2) ! strcmp((s1),(s2))

//static const int MAXSTRINGSIZE = 255;

/*
 * This is ! NULL iff I am presently taking commands from a file,
 * (not stdin)
 */
static FILE *logfile = NULL;

/*
 * Initialize logging
 */
extern void
open_log()
{
  if (logfile) {
    gnans_error(non_fatal, "open_log called, but logfile != NULL");
    return;
  }
  if (! (logfile = fopen(GNANSLOGFILENAME, "w"))) {
    gnans_error(non_fatal, "Could not open file ``%s''", GNANSLOGFILENAME);
    return;
  }

  fprintf(logfile, "Log from Gnans version %s,\n", version_string);
  fprintf(logfile, "compiled on %s,\nconfigured with command: %s.\n",
	  compiled_string, compilation_characteristics);
  fprintf(logfile, "Created by %s@%s.\n", cuserid(NULL),
	  gethostname_as_string().c_str());
  fprintf(logfile, "Starting at %s.\n", getdate());
  fflush(logfile);
}

/*
 * Finish logging
 */
extern void
close_log()
{
  if (! logfile)
    return;
  log_event("Log file closed");
  if (fclose(logfile))
    gnans_error(non_fatal, "Cannot close file %s", GNANSLOGFILENAME);
  logfile = NULL;
}

/*
 * Initialize OR finish logging
 */
extern void
update_log_status()
{
  if (do_logging)
    open_log();
  else
    close_log();
}

/*
 * Enter a its argument into the log.
 * If called with an empty string, just flushes the log.
 */
extern void
log_event(const char *str)
{
  if (! logfile)
    return;
  if (STR_EQUAL(str, "")) {
    fflush(logfile);
    return;
  }
  if (running_script && ! log_script_commands)
    return;
  if (fprintf(logfile, "[%s]: %s\n", getdate(), str) == EOF)
    gnans_error(non_fatal,
		"Error occurred writing to logfile.\nString = ``%s''", str);
  if (! running_script)
    fflush(logfile);
}

/*
 * Variable argument form of log_event. Arguments are like printf.
 */
extern void
va_log_event(const char *fmt,...)
{
  char log_str[10*MAXSTRINGSIZE];
  va_list args;
  va_start(args, fmt);
  vsprintf(log_str, fmt, args);
  va_end(args);
  log_event(log_str);
}

/*
 * Like display_vars, but writes the result in the logfile
 */
extern void
display_in_logfile()
{
  if (! logfile)
    return;
  char *str = create_display_string(0);
  fprintf(logfile, "*************** Display ***************\n");
  fprintf(logfile, "%s", str);
  fprintf(logfile, "******************************\n");
  delete [] str;
}
