/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: gnanslator.cc,v 1.10 1994/11/01 12:53:43 bengt Exp bengt $ */

/*
 * This is the main file for the translator.
 */

#include <config.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <utillib.h>
#include "translator.h"

extern "C" char * cuserid ( char *string ); // Flabeg
/* My name */
static std::string progname;

/* Forwards */
static std::string format_eqn(symrec *symb, bool protect_wiener_coeffs);
extern int main(int argc, char *argv[]);

/* 
 * This table contains all reserved words in ANSI C, in order to barf
 * on such already when translating.
 * (Source: Stroustrup, Second edition)
 */
static char *keywords[] = {
  "asm",
  "auto",
  "break",
  "case",
  "char",
  "class",
  "const",
  "continue",
  "default",
  "delete",
  "do",
  "double",
  "else",
  "enum",
  "extern",
  "float",
  "for",
  "friend",
  "goto",
  "if",
  "inline",
  "int",
  "long",
  "new",
  "operator",
  "private",
  "protected",
  "public",
  "register",
  "return",
  "short",
  "signed",
  "sizeof",
  "static",
  "struct",
  "switch",
  "template",
  "this",
  "throw",
  "try",
  "typedef",
  "union",
  "unsigned",
  "virtual",
  "void",
  "volatile",
  "while",
  ""};

static int no_option_args_kludge = 0;

/***************** Open all files *****************/
static void 
open_it(int argc, char *argv[])
{
  if (argc > 1) {
    FILE *f;
    f = fopen(argv[1], "r");
    if (f)
      input_filename = argv[1];
    else {
      input_filename = std::string(argv[1]) + ".gnans";
      f = fopen(input_filename.c_str(), "r");
      if (!f)
	gnans_error(fatal, "Cannot open input file %s[.gnans]\n", argv[1]);
    }
    zzin = f;
  } else {
    zzin = stdin;
    input_filename = "STDIN";
  }

  if (output_filename == "")
    if (input_filename.find(".") != std::string::npos)//FIXME
      output_filename = input_filename.substr(0, input_filename.find_last_of(".")) + OUTPUT_ENDING;
    else
      output_filename = input_filename + OUTPUT_ENDING;
  
  if (!(outputfile = fopen(output_filename.c_str(),"w")))
    gnans_error(fatal, "Cannot open `%s' for writing",
		output_filename.c_str());
  zzout = outputfile;
}

/* Write the preamble of the output file, basically static info */
static void 
preamble(unsigned int argc, char *argv[])
{
  unsigned int i;

  fprintf(outputfile,"/* Generated by %s dated %s */\n",
	  progname.c_str(), version_string);
  fprintf(outputfile,"/* at %s for %s@%s */\n",
	  getdate(), cuserid(0), gethostname_as_string().c_str());
  fprintf(outputfile,"/* from command `");
  for (i = 0; i < argc + no_option_args_kludge; i++) {
    fprintf(outputfile, "%s ", argv[i - no_option_args_kludge]);
  }
  fprintf(outputfile,"' */\n");
  fprintf(outputfile,
	  "#ifdef HAVE_CONFIG_H\n"
	  "#include <config.h>\n"
	  "#endif\n"
	  "#include <math.h>\n"
	  "#include <string.h>\n"
	  "#include <stdlib.h>\n"
	  "#include <time.h>\n"
	  "#define IS_GNANS_SYSTEM\n"
	  "#include \"externals.h\"\n"
	  "#include \"gnanslib.h\"\n"
	  "#include \"random.h\"\n\n"
	  "static bool do_stochastics;\n\n");
}

/* Output the C-DECLARATIONS */
static void 
write_c_declarations()
{
  if (c_declarations != "")
    fprintf(zzout, "/* C declarations: */\n%s", c_declarations.c_str());
}

/* Output the data structures holding all the user parameters */
static void 
write_arrays()
{
  unsigned int i;
  fprintf(zzout, "  delete [] current_system.parameters;\n"
	  "  current_system.parameters = NULL;\n");
  if (number_of_parameters)
    fprintf(zzout,
	    "  current_system.parameters = new UserParameterPointer[%d];\n",
	    number_of_parameters);
  for (i = 0; i < number_of_parameters; i++) {
    fprintf(zzout,
	    "  gnans_parameter_%s.init(\"%s\", \"%s\", false, %0.16e);\n",
	    (char *) parameters[i], (char *) parameters[i],
	    parameters[i].documentation, parameters[i].initial);
    fprintf(zzout,
	    "  current_system.parameters[%d] = &gnans_parameter_%s;\n",
	    i, (char *) parameters[i]);
  }

  if (number_of_parameter_matrices)
    fprintf(zzout, "  unsigned int params_this_far = 0;\n\n");
  for (i = 0; i < number_of_parameter_matrices; i++) {
    fprintf(zzout,
	    "  success = success &&\n"
	    "    gnans_parameter_matrix_%s.matrix_init(\"%s\", \"%s\", "
	    "false, \"%s\", \"%s\", "
	    "current_system.parameters + params_this_far);\n"
	    "  params_this_far += gnans_parameter_matrix_%s.rows() * gnans_parameter_matrix_%s.cols();\n",
	    (char *) parameter_matrices[i], (char *) parameter_matrices[i],
	    parameter_matrices[i].documentation,
	    parameter_matrices[i].filename,
	    parameter_matrices[i].external_name,
	    (char *) parameter_matrices[i], (char *) parameter_matrices[i]);
  }
  
  fprintf(zzout, "  delete [] current_system.functions;\n"
	  "  current_system.functions = NULL;\n");
  if (number_of_functions)
    fprintf(zzout,
	    "  current_system.functions = new UserParameterPointer[%d];\n",
	    number_of_functions);
  for (i = 0; i < number_of_functions; i++) {
    fprintf(zzout,
	    "  gnans_function_%s.init(\"%s\", \"%s\", store_as_default);\n",
	    (char *) functions[i], (char *) functions[i],
	    functions[i].documentation);
    fprintf(zzout,
	    "  current_system.functions[%d] = &gnans_function_%s;\n",
	    i, (char *) functions[i]);
  }

  fprintf(zzout, "  delete [] current_system.randoms;\n"
	  "  current_system.randoms = NULL;\n");
  if (number_of_randoms)
    fprintf(zzout,
	    "  current_system.randoms = new UserParameterPointer[%d];\n",
	  number_of_randoms);
  for (i = 0; i < number_of_randoms; i++) {
    fprintf(zzout,
	    "  gnans_random_%s.init(\"%s\", \"%s\", false, %g);\n",
	    (char *) randoms[i], (char *) randoms[i],
	    randoms[i].documentation, randoms[i].initial);
    fprintf(zzout,
	    "  current_system.randoms[%d] = &gnans_random_%s;\n"
	    "  gnans_random_%s.rng_ptr = new rng;\n"
	    "  gnans_random_%s.rng_ptr->save_state(0);\n",
	    i, (char *) randoms[i], (char *) randoms[i], (char *) randoms[i]);
  }
  
  fprintf(zzout, "  delete [] current_system.wieners;\n"
	  "  current_system.wieners = NULL;\n");
  if (number_of_wieners)
    fprintf(zzout,
	    "  current_system.wieners = new UserParameterPointer[%d];\n"
	    "  current_system.wiener_coeffs = new Vect[%d];\n",
	    number_of_wieners, number_of_wieners);
  for (i = 0; i < number_of_wieners; i++)
    fprintf(zzout,
	    "  current_system.wieners[%d] = &gnans_random_%s;\n"
	    "  current_system.wiener_coeffs[%d].resize(%d);\n",
	    i, (char *) wieners[i], i, number_of_states);
  
  fprintf(zzout, "  delete [] current_system.states;\n"
	  "  current_system.states = NULL;\n");
  if (number_of_states)
    fprintf(zzout,
	    "  current_system.states = new UserParameterPointer[%d];\n",
	    number_of_states);
  for (i = 0; i < number_of_states; i++) {
    fprintf(zzout,
	    "  gnans_state_%s.init(\"%s\", \"%s\", store_as_default, %0.16e);\n",
	    (char *) states[i], (char *) states[i],
	    states[i].documentation, states[i].initial);
    fprintf(zzout,
	    "  current_system.states[%d] = &gnans_state_%s;\n",
	    i, (char *) states[i]);
  }
  
  for (i = 0; i < number_of_state_matrices; i++) {
    fprintf(zzout,
	    "  success = success &&\n"
	    "    gnans_state_matrix_%s.matrix_init(\"%s\", \"%s\", store_as_default, \"%s\", \"%s\", current_system.states);\n",
	    (char *) state_matrices[i], (char *) state_matrices[i],
	    state_matrices[i].documentation,
	    state_matrices[i].filename,
	    state_matrices[i].external_name);
    // Bind in the state components ...
    //fprintf(zzout,
    //    "  current_system.states[%d] = &gnans_state_%s;\n",
    //  i, (char *) states[i]);
  }
  fprintf(zzout,
	  "  if (!success)\n"
	  "    return false;\n");
}

/* Output the CPP wizardry */
static void 
cpp_stuff()
{
  unsigned int i, j;
  for (i = 0; i < number_of_states; i++)
    fprintf(zzout, "#define %s state_vector[%d]\n", (char *) states[i], i);
  if (system_type != discrete)
    for (i = 0; i < number_of_states; i++)
      fprintf(zzout, "#define gnans_d_%s gnans_d_states[%d]\n",
	      (char *) states[i],i);
  for (j = 0; j < number_of_wieners; j++)
    for (i = 0; i < number_of_states; i++)
      fprintf(zzout,
	      "#define %s_%s ((double*)current_system.wiener_coeffs[%d])[%d]\n",
	      (char *) states[i], (char *) wieners[j], j, i);
  for (i = 0; i < number_of_functions; i++)
    fprintf(zzout, "#define %s gnans_function_%s\n",
	    (char *) functions[i], (char *) functions[i]);
  for (i = 0; i < number_of_parameters; i++)
    fprintf(zzout, "#define %s gnans_parameter_%s\n",
	    (char *) parameters[i], (char *) parameters[i]);
  for (i = 0; i < number_of_variables; i++)
    fprintf(zzout, "#define %s gnans_variable_%s\n",
	    (char *) variables[i], (char *) variables[i]);
  for (i = 0; i < number_of_uniforms; i++)
    fprintf(zzout, "#define %s gnans_random_%s_localvalue\n",
	    (char *) uniforms[i], (char *) uniforms[i]);
  for (i = 0; i < number_of_normals; i++)
    fprintf(zzout, "#define %s gnans_random_%s_localvalue\n",
	    (char *) normals[i], (char *) normals[i]);
  fprintf(zzout, "#define %s *time_ptr\n", time_variable_name.c_str());
  fprintf(zzout, "\n");
}

static void
write_name()
{
  fprintf(zzout,
	  "\n"
	  "static double *state_vector;\n"
	  "static double *time_ptr;\n\n");
  if (system_type == riccati)
    fprintf(zzout, "\nstatic double *riccati_problem_matrix;\n\n");
  fprintf(zzout,
	  "static char *this_system_name = \"%s\";\n"
	  "static char *this_system_time_name = \"%s\";\n\n",
	  system_name.c_str(), time_variable_name.c_str());
}

static void
write_riccati_setup()
{
  fprintf(zzout,
	  "  unsigned int nnp = gnans_state_matrix_%s.rows();\n"
	  "  unsigned int nnq = gnans_state_matrix_%s.cols();\n",
	  (char *) state_matrices[0],
	  (char *) state_matrices[0]);

  fprintf(zzout,
	  "  if (!(\n"
	  "          gnans_parameter_matrix_%s.rows() == nnq &&\n"
	  "          gnans_parameter_matrix_%s.rows() == nnq &&\n"
	  "          gnans_parameter_matrix_%s.rows() == nnp &&\n"
	  "          gnans_parameter_matrix_%s.rows() == nnp &&\n"
	  "          gnans_parameter_matrix_%s.cols() == nnq &&\n"
	  "          gnans_parameter_matrix_%s.cols() == nnp &&\n"
	  "          gnans_parameter_matrix_%s.cols() == nnq &&\n"
          "          gnans_parameter_matrix_%s.cols() == nnp)) {\n"
	  "     gnans_error(non_fatal, \"Incompatible dimensions\");\n"
	  "     return false;\n"
	  "  }\n",
	  riccati_data.A11->name,
	  riccati_data.A12->name,
	  riccati_data.A21->name,
	  riccati_data.A22->name,
	  riccati_data.A11->name,
	  riccati_data.A12->name,
	  riccati_data.A21->name,
	  riccati_data.A22->name);

#if 0
  fprintf(zzout,
	  "  double *Ap = riccati_problem_matrix = new double[sqr(nnp+nnq)];\n"
	  "  double *A11p = gnans_parameter_matrix_%s;\n"
	  "  double *A12p = gnans_parameter_matrix_%s;\n"
	  "  double *A21p = gnans_parameter_matrix_%s;\n"
	  "  double *A22p = gnans_parameter_matrix_%s;\n"
	  "  unsigned int j;\n"
	  "  for (unsigned int i = 0; i < nnq; i++) {\n"
	  "    for (j = 0; j < nnq; j++) {\n"
	  "      *Ap++ = *A11p++;\n"
	  "    }\n"
	  "    for (j = 0; j < nnp; j++) {\n"
	  "      *Ap++ = *A21p++;\n"
	  "    }\n"
	  "  }\n"
	  "  for (i = 0; i < nnp; i++) {\n"
	  "    for (j = 0; j < nnq; j++) {\n"
	  "      *Ap++ = *A12p++;\n"
	  "    }\n"
	  "    for (j = 0; j < nnp; j++) {\n"
	  "      *Ap++ = *A22p++;\n"
	  "    }\n"
	  "  }\n",
	  riccati_data.A11->name,
	  riccati_data.A12->name,
	  riccati_data.A21->name,
	  riccati_data.A22->name);
#endif
  fprintf(zzout, "  riccati_problem_matrix = new double[sqr(nnp+nnq)];\n");
}

/* Generate the setup function, to be called once at system initialization */
static void 
write_setup()
{
  unsigned int i;
  fprintf(zzout,
	  "extern \"C\" int\n"
	  "gnans_setup(double **gnans_state_vector_ptr, double *system_time)\n"
	  "{\n"
	  "  bool success = true;\n\n");
  
  if (system_type != riccati)
    fprintf(zzout,
	    "  state_vector = new double[%d];\n"
	    "  current_system.number_of_states = %d;\n",
	    number_of_states ? number_of_states : 1, number_of_states);

  write_arrays();
  
  if (system_type == riccati) {
    write_riccati_setup();
    fprintf(zzout,
	    "  state_vector = new double[gnans_state_matrix_%s.cols() * \n"
	    "                            gnans_state_matrix_%s.rows()];\n"
	    "  for (unsigned int i = 0; i < gnans_state_matrix_%s.cols() * gnans_state_matrix_%s.rows(); i++)\n"
	    "      state_vector[i] = *((double *)gnans_state_matrix_%s + i);\n"
	    "  current_system.number_of_states = gnans_state_matrix_%s.cols() * \n"
	    "                            gnans_state_matrix_%s.rows();\n"
	    "  current_system.state_rows = gnans_state_matrix_%s.rows();\n"
	    "  current_system.state_cols = gnans_state_matrix_%s.cols();\n",
	    (char *) state_matrices[0], (char *) state_matrices[0],
	    (char *) state_matrices[0], (char *) state_matrices[0],
	    (char *) state_matrices[0],
	    (char *) state_matrices[0], (char *) state_matrices[0],
	    (char *) state_matrices[0], (char *) state_matrices[0]);

  } else
      for (i = 0; i < number_of_states; i++)
	fprintf(zzout,
		"  %s = gnans_state_%s;\n",
		(char *) states[i], (char *) states[i]);

  fprintf(zzout,
	  "  current_system.number_of_parameters = %d;\n",
	  number_of_parameters);
  for (i = 0; i < number_of_parameter_matrices; i++)
    fprintf(zzout,
	    "  current_system.number_of_parameters +=\n"
	    "    gnans_parameter_matrix_%s.rows() * gnans_parameter_matrix_%s.cols();\n",
	    (char *) parameter_matrices[i], (char *) parameter_matrices[i]);
  fprintf(zzout,
	  "  current_system.number_of_functions = %d;\n"
	  "  current_system.number_of_randoms = %d;\n"
	  "  current_system.number_of_uniforms = %d;\n"
	  "  current_system.number_of_normals = %d;\n"
	  "  current_system.number_of_wieners = %d;\n"
	  "  current_system.system_type = (SystemType) %d;\n",
	  number_of_functions, number_of_randoms, number_of_uniforms,
	  number_of_normals, number_of_wieners, system_type);
  
  fprintf(zzout,
	  "  *gnans_state_vector_ptr = state_vector;\n"
	  "  time_ptr = system_time;\n");
  fprintf(zzout, "  current_system.system_name = this_system_name;\n");
  fprintf(zzout, "  systemtime.setname(this_system_time_name);\n"
	  "  return success;\n}\n\n");
}

/****************************************************************/
/* All the sorting stuff follows */

/*
 * This function is for debugging:
 * dump the symbol table and dependencies onto stderr.
 */
static void 
print_deps(symrec *start)
{
  symrec *p;
  deprec *d;
  fprintf(stderr, "**************** Initial symbol table:\n");
  for (p = start; p; p = p->next) {
    fprintf(stderr, "`%s' depends on:", format_eqn(p, false).c_str());
    for (d = p->dependencies; d; d = d->next) {
      if (d->symbol)
	fprintf(stderr, " %s", d->symbol->name);
      else
	fprintf(stderr, " Empty");
    }    
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "****************\n");
}

// Remove parameters, c_identifiers, and keywords from the symbol table
static void 
zap_parameters()
{
  symtab_delete_type(Parameter);
  symtab_delete_type(Parameter_vector);
  symtab_delete_type(Parameter_matrix);
  symtab_delete_type(C_Identifier);
  symtab_delete_type(Keyword);
  symtab_delete_type(Wiener_Process); // !!
}

/* Remove times etc from the symbol table */
static void 
zap_states_and_times()
{
  unsigned int i;
  symtab_delete_type(Time_type);
  symtab_delete_type(Uniform_Noise);
  symtab_delete_type(Normal_Noise);
  // symtab_delete_type(Wiener_Process);
  symtab_delete_type(Integer_State); /* Don't know what to do with these */
  for (i = 0; i < number_of_states; i++)
    dependency_delete(states[i]);
  for (i = 0; i < number_of_state_vectors; i++)
    dependency_delete(state_vectors[i]);
  for (i = 0; i < number_of_state_matrices; i++)
    dependency_delete(state_matrices[i]);
}

/* Returns a text string containing the defining equation for SYMB */
static std::string
format_eqn(symrec *symb, bool protect_wiener_coeffs)
{
  std::string str;
  if (symb->equation) {
    str = std::string("#line ") + number_String(symb->line_no_of_equation) + " \"" + input_filename + "\"\n  ";
    if (protect_wiener_coeffs && symb->type == Wiener_coeff)
      str += "if (do_stochastics) ";
    if (system_type != riccati) {
      if (symb->type == State && system_type == continuous)
	str += "gnans_d_";
      if (symb->type == State && system_type == discrete)
	str += "double gnans_q_";
      str += symb->name;
      // I have to cast RHS to double to avoid problems when RHS
      // is a UserParameter.
      str += "=double(";
    }
    str += symb->equation + std::string(");");
  } else {
    str = std::string(symb->name) + ": *** No equation ***";
  }
  return str;
}

/*
 * This is the basic function for the sorting of the equations.
 * Checks if THIS_SYM is ready for outputting, i.e. if it doesn't depending on
 * anything unresolved. If it depends on something unresolved, calls itself
 * recursively with the dependent as argument and the calling symbol as parent.

 * Third argument states if the time is to be considered as resolved or not.
 */
static bool 
investigate(symrec *this_sym, symrec *parent, bool timeinvariant,
	    unsigned int level = 1)
{
  symrec *ptr;
  deprec *dptr;
  bool success = true;
  if (! this_sym) {
    if (deps_debug)
      fprintf(stderr, "Investigate called with null argument, returning\n");
    return true;
  }
  if (deps_debug) {
    fprintf(stderr, "[%d](%s) Investigate %s: ", level,
	    timeinvariant ? "Timeinvariant" : "Time varying", this_sym->name);
    if (parent)
      fprintf(stderr, "Parent =  %s", parent->name);
    fprintf(stderr, "\n");
  }
  if (timeinvariant && ( this_sym->type == State
			|| this_sym->type == State_vector
			|| this_sym->type == State_matrix
			|| this_sym->type == Integer_State
			|| this_sym->type == Wiener_coeff
			|| this_sym->type == Time_type
			|| this_sym->type == Uniform_Noise
			|| this_sym->type == Normal_Noise
			|| this_sym->type == Wiener_Process )) {
    if (deps_debug)
      fprintf(stderr, "%s not timeinvariant -- punting\n",
	      format_eqn(this_sym, false).c_str());
    return false;
  }

  // If parent is non-NULL, this is an algebraic loop
  if (this_sym->parent) {
    fprintf(stderr, "Algebraic loop:\n");
    // Report the loop in the order we went through
    for (unsigned int i = level - 1; i >= 1; i--) {
      ptr = parent;
      for (unsigned int j = 1; j < i; j++) {
	ptr = ptr->parent;
      }
      fprintf(stderr, "  %s\n", format_eqn(ptr, false).c_str());
    }
    gnans_error(fatal, "");
  }

  // If this_sym does not depend on anything unresolved, we are finished...
  if (!this_sym->dependencies) {
    if (!this_sym->equation) {
      gnans_error(recoverable, "`%s' has no equation", this_sym->name);
      return true;
    }
    fprintf(zzout, "%s\n", format_eqn(this_sym, true).c_str());
    symtab_delete(this_sym->name);
    if(deps_debug)
      fprintf(stderr, "`%s' successfully eliminated!\n",
	      format_eqn(this_sym, false).c_str());
    return true;
  } else {
    // ...otherwise try to process what this_sym depends on.
    this_sym->parent = parent ? parent : this_sym;
    for (dptr = this_sym->dependencies; dptr; dptr = dptr->next)
      success = success &&
	investigate(dptr->symbol, this_sym, timeinvariant, level + 1);
    this_sym->parent = NULL;
    if (success) {
      fprintf(zzout, "%s\n", format_eqn(this_sym, true).c_str());
      symtab_delete(this_sym->name);
      if(deps_debug)
	fprintf(stderr, "%s successfully eliminated!!\n",
		format_eqn(this_sym, false).c_str());
      return true;
    }
    return false;
  }
}

/* Output all timeinvariant equations ("AT TIME START:") */
static void 
time_invariant_equations()
{
  symrec *s;
  /* zap_parameters(); */
  for (s = sym_table; s; s = s->next)
    investigate(s, NULL, true);

  /* Now output those guys explicitly in the AT TIME START section */
  for (s = sym_table; s; s = s->next)
    if (s->a_priori_timeindex == STARTINDEX) {
      fprintf(zzout, "%s\n", format_eqn(s, false).c_str());
      symtab_delete(s->name);
      if(deps_debug)
	fprintf(stderr, "%s eliminated as a priori time invariant\n",
		format_eqn(s ,false).c_str());
    }
}

/* Output all dynamic equations (the equations necessary for state updating).*/
static void 
dynamic_equations()
{
  unsigned int i, j;
  bool success = true;
  if (deps_debug)
    fprintf(stderr, "**************** Now doing the dynamical equations --\n");
  zap_states_and_times();

  for (i = 0; i < number_of_states; i++)
    for (j = 0; j < number_of_wieners; j++)
      success = success &&
	investigate(getsym(stringify("%s_%s",
				     (char *) states[i],
				     (char *) wieners[j])),
		    NULL, false);

  for (i = 0; i < number_of_states; i++)
    success = success && investigate(getsym(states[i]), NULL, false);
  for (i = 0; i < number_of_state_vectors; i++)
    success = success && investigate(getsym(state_vectors[i]), NULL, false);
  for (i = 0; i < number_of_state_matrices; i++)
    success = success && investigate(getsym(state_matrices[i]), NULL, false);
  if (! success)
    gnans_error(impossible, "Internal error: did not manage to sort all states!");
}
  
/*
 * Output all output equations
 * (not timeinvariant, not necessary for updatingh the state).
 */
static void 
output_equations()
{
  symrec *s;
  bool success = true;
  if (deps_debug)
    fprintf(stderr, "**************** Now doing the output equations --\n");
  for (s = sym_table; s; s = s->next)
    success = success && investigate(s, NULL, false);
  if (!success)
    gnans_error(impossible, "Internal error: did not manage to sort all symbols");
}

/**************** Define some static storage for rng's ****************/
static void
write_random_storage()
{
  unsigned int i;
  for (i = 0; i < number_of_uniforms; i++)
    fprintf(zzout, "static double gnans_random_%s_localvalue;\n\n",
	    (char *) uniforms[i]);
  for (i = 0; i < number_of_normals; i++)
    fprintf(zzout, "static double gnans_random_%s_localvalue;\n\n",
	    (char *) normals[i]);
}

/* Output the init function, to be called once for each simulation run. */
static void 
write_init()
{
  unsigned int i;
  fprintf(zzout,
	  "extern \"C\" int\n"
	  "gnans_init(double gnans_state_vector[], Continuation continuation, bool stochastic)\n"
	  "{\n"
	  "  do_stochastics = stochastic;\n");
  fprintf(zzout, "  if (continuation == init_none)\n"
	  "    return true;\n\n");

  if (system_type == riccati) {
    fprintf(zzout,
	    "  double *Ap = riccati_problem_matrix;\n"
	    "  double *A11p = gnans_parameter_matrix_%s;\n"
	    "  double *A12p = gnans_parameter_matrix_%s;\n"
	    "  double *A21p = gnans_parameter_matrix_%s;\n"
	    "  double *A22p = gnans_parameter_matrix_%s;\n"
	    "  unsigned int i, j;\n"
	    "  for (i = 0; i < gnans_state_matrix_%s.cols(); i++) {\n"
	    "    for (j = 0; j < gnans_state_matrix_%s.cols(); j++) {\n"
	    "      *Ap++ = *A11p++;\n"
	    "    }\n"
	    "    for (j = 0; j < gnans_state_matrix_%s.rows(); j++) {\n"
	    "      *Ap++ = *A21p++;\n"
	    "    }\n"
	    "  }\n"
	    "  for (i = 0; i < gnans_state_matrix_%s.rows(); i++) {\n"
	    "    for (j = 0; j < gnans_state_matrix_%s.cols(); j++) {\n"
	    "      *Ap++ = *A12p++;\n"
	    "    }\n"
	    "    for (j = 0; j < gnans_state_matrix_%s.rows(); j++) {\n"
	    "      *Ap++ = *A22p++;\n"
	    "    }\n"
	    "  }\n",
	    riccati_data.A11->name,
	    riccati_data.A12->name,
	    riccati_data.A21->name,
	    riccati_data.A22->name,
	    (char *) state_matrices[0],
	    (char *) state_matrices[0],
	    (char *) state_matrices[0],
	    (char *) state_matrices[0],
	    (char *) state_matrices[0],
	    (char *) state_matrices[0]);

    fprintf(zzout,
	   "  for (i = 0; i < current_system.number_of_states; i++)\n"
	   "    state_vector[i] = *((double *) gnans_state_matrix_%s + i);\n",
	   (char *) state_matrices[0]);
  } else
    for (i = 0; i < number_of_states; i++)
      fprintf(zzout,
	      "  %s = gnans_state_%s;\n",
	      (char *) states[i], (char *) states[i]);
  
  fprintf(zzout,
	  "  if (continuation == init_states)\n"
	  "    return true;\n\n");

  for (i = 0; i < number_of_randoms; i++) {
    fprintf(zzout,
	    "  gnans_random_%s.rng_ptr->restore_state(0);\n"
	    "  gnans_random_%s.rng_ptr->srandom(int(gnans_random_%s) ?\n"
	    "                               int(gnans_random_%s) \n"
	    "                              : time(NULL));\n",
	    (char *) randoms[i], (char *) randoms[i],
	    (char *) randoms[i], (char *) randoms[i]);
  }

  fprintf(zzout, "%s\n", times_buffer[STARTINDEX].c_str());
  time_invariant_equations();
  fprintf(zzout,"  return true;\n}\n\n");
}

/* Output the state updating function */
static void 
write_f(bool /*hairy*/)
{
  switch (system_type) {
  case discrete:
    fprintf(zzout,
	    "extern \"C\" void\n"
	    "gnans_discrete_f()\n{\n");
    break;
  case riccati:
    fprintf(zzout,
	    "extern \"C\" void\n"
	    "gnans_f(double *time_ptr, double *A,\n"
            "        double *rarr /*, int *iarr*/)\n"
	    "{\n");
    break;
  default:
    fprintf(zzout,
	    "extern \"C\" void\n"
	    "gnans_f(double *time_ptr,\n"
	    "\tdouble state_vector[%d],\n"
	    "\tdouble gnans_d_states[%d])\n{\n",
	    number_of_states,
	    number_of_states);
  }
  fprintf(zzout,
	  "#ifdef LOGCALLS\n"
	  "  va_log_event(\"f called at time %%g\", *time_ptr);\n"
	  "#endif\n");

  /* Give all random things their local values */
  unsigned int i;
  for (i = 0; i < number_of_normals; i++)
    fprintf(zzout,
	    "  %s = gnans_random_%s.rng_ptr->normal();\n",
	    (char *) normals[i], (char *) normals[i]);
  for (i = 0; i < number_of_uniforms; i++)
    fprintf(zzout,
	    "  %s = gnans_random_%s.rng_ptr->uniform();\n",
	    (char *) uniforms[i], (char *) uniforms[i]);
  
  for (i = ACCEPTEDINDEX + 1; i < number_of_times; i++)
    if (times_names[i] == time_variable_name)
      break;
  
  if (i == number_of_times)
    gnans_error(fatal, "No system time defined");

  if (system_type == riccati)
    fprintf(zzout, "\n"
	    "  bcopy((char*) riccati_problem_matrix,\n"
	    "        (char*) A,\n"
	    "        sizeof(double)*sqr(current_system.state_cols\n"
	    "                          + current_system.state_rows));\n");
  if (translator_debug)
    fprintf(zzout, "  /* Now doing times_buffer[%d] */\n", i);
  fprintf(zzout, "%s\n", times_buffer[i].c_str());
  if (translator_debug)
    fprintf(zzout, "  /* Finished with times_buffer[%d] */\n", i);
  dynamic_equations();
  if (system_type == discrete) {
    //fprintf(zzout, "  /* Updating of the state */\n");
    for (i = 0; i < number_of_states; i++)
      fprintf(zzout, "  %s = gnans_q_%s;\n",
	      (char *) states[i], (char *) states[i]);
  }
  fprintf(zzout, "}\n\n");
}

/* Output the accepted function */
static void 
write_f_accepted()
{
  fprintf(zzout,
	  "extern \"C\" void\n"
	  "gnans_f_accepted()\n{\n");
  fprintf(zzout, "%s", times_buffer[ACCEPTEDINDEX].c_str());
  if (times_buffer[ACCEPTEDINDEX] != "")
    fprintf(zzout, "\n");
  fprintf(zzout, "}\n\n");
}

/* Output the output equation function */
static void 
write_f_store()
{
  fprintf(zzout,
	  "extern \"C\" void\n"
	  "gnans_f_store()\n{\n");
  fprintf(zzout, "%s", times_buffer[STOREINDEX].c_str());
  if (times_buffer[STOREINDEX] != "")
    fprintf(zzout, "\n");
  output_equations();
  fprintf(zzout, "}\n\n");
}

/* Output the ending function ("AT TIME STOP:") */
static void 
write_f_finish()
{
  fprintf(zzout,
	  "extern \"C\" void\n"
	  "gnans_finish()\n{\n");
  fprintf(zzout, "%s", times_buffer[STOPINDEX].c_str());
  fprintf(zzout, "}\n\n");
}

/* Initialize the symtable */
static void 
setup_symtable()
{
  char **p;
  
  putsym("START", Time_type);
  putsym("STOP", Time_type);
  putsym("STORE", Time_type);
  putsym("ACCEPTED", Time_type);
  for (p = keywords; **p != '\0'; p++)
    putsym(*p, Keyword);
}

static void
write_dummy()
{
  if (system_type == discrete)
    fprintf(zzout,
	    "extern \"C\" void\n"
	    "gnans_f(double *gnans_t,\n"
	    "\tdouble gnans_state_vector[],\n"
	    "\tdouble gnans_d_states[])\n{\n"
	    "  gnans_error(impossible, \"This can't happen\");\n"
	    "}\n\n");
  else
    fprintf(zzout,
	    "extern \"C\" void\n"
	    "gnans_discrete_f()\n{\n"
	    "  gnans_error(impossible, \"This can't happen\");\n"
	    "}\n\n");
}

static void 
write_destroy()
{
  fprintf(zzout,
	  "extern \"C\" void\n"
	  "gnans_destroy()\n"
	  "{\n"
	  "  delete [] state_vector;\n"
	  "  state_vector = NULL;\n");
  for (unsigned int i = 0; i < number_of_randoms; i++)
    fprintf(zzout,
	    "  delete gnans_random_%s.rng_ptr;\n",
	    (char *) randoms[i]);
  if (number_of_wieners)
    fprintf(zzout, "  delete [] current_system.wiener_coeffs;\n"
	    "  current_system.wiener_coeffs = NULL;\n");
  if (system_type == riccati )
    fprintf(zzout, "  delete [] riccati_problem_matrix;\n"
	    "  riccati_problem_matrix = NULL;\n");
  fprintf(zzout, "}\n");
}

extern int 
main(int argc, char *argv[])
{
  unsigned int i;
  char *START = "START";
  char *STOP  = "STOP";
  char *STORE = "STORE";
  char *ACCEPTED = "ACCEPTED";
  bool raw_code;
  zz_flex_debug = 0;
  progname = basename(argv[0]);
  if (argc >= 3)
    if (strcmp(argv[1],"-debug") == 0) {
      sscanf(argv[2], "%d", &translator_debug);
      zz_flex_debug = translator_debug & 1; /* Scanner debugging */
      zzdebug = translator_debug & 2;	    /* Parser debugging */
      deps_debug = translator_debug & 4;    /* Dependencies debugging */
      trace_lineno = translator_debug & 8;
      trace_mallocs = translator_debug & 16;
      argc--; argv++;
      argc--; argv++;
      no_option_args_kludge += 2;
    }

  output_filename = "";
  if (argc >= 3)
    if (strcmp(argv[1], "-o") == 0) {
      output_filename = argv[2];
      argc--; argv++;
      argc--; argv++;
      no_option_args_kludge += 2;
    }
  if (argc > 2)
    gnans_error(fatal, "Usage: gnans-trans [-debug <value>] [-o outputfile] [file[.gnans]]");

  time_variable_name = "__undefined__";
  c_declarations = "";
  times_names[STARTINDEX] = START;
  times_names[STOPINDEX] = STOP;
  times_names[STOREINDEX] = STORE;
  times_names[ACCEPTEDINDEX] = ACCEPTED;
  for (i = 0; i <= ACCEPTEDINDEX; i++)
    times_buffer[i] = "";
  open_it(argc, argv);
  preamble(argc, argv);
  setup_symtable();
  zzparse();
  symtab_delete_type(Defun);
  assignment_check();
  if ((system_type == continuous || system_type == hybrid)
      && number_of_states == 0)
    gnans_error(fatal, "You have to have at least one dynamic equation.");
  if (errors_occured)
    gnans_error(fatal, "Quitting due to errors.");
  write_name();
  write_random_storage();
  cpp_stuff();
  write_c_declarations();
  zap_parameters();
  if (deps_debug)
    print_deps(sym_table);
  write_setup();
  write_init();
  write_f(false);
  // This doesn't work presently;
  // dynamic_equation cannot be called more than once
  // write_f(true); 
  write_f_accepted();
  write_f_store();
  write_f_finish();
  write_dummy();
  write_destroy();
  if (fclose(outputfile))
    gnans_error(fatal, "Cannot close file `%s'", output_filename.c_str());

  // Report to the user how deeply we have checked his/her code
  raw_code = c_declarations != "";
  for (i = 0; i < number_of_times; i++)
    raw_code = raw_code || (times_buffer[i] != "");
  if (raw_code || cpp_calls)
    fprintf(stderr,
	    "Translation succeeded with UNSAFE status.\n"
	    "I can't guaranteed success during compilation or linking due to\n"
	    "raw C code and/or function calls.\n");
  else
    if (! function_calls)
      fprintf(stderr, "Translation succeeded with SAFE status.\n"
	      "No errors should occur during compilation or linking.\n");
    else
      fprintf(stderr, "Translation succeeded  with SEMI-SAFE status.\n"
	      "No errors should occur during compilation.\n"
	      "However, I can't guaranteed success during linking due to function calls.\n"
	      "(Watch out for compile warnings concering implicitly declared functions.)\n");
  return(0);
}
