%{
/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: parser.y,v 1.30 1994/11/01 12:53:43 bengt Exp bengt $ */

/* Debugging support enabled -- also in production versions */
#define YYDEBUG 1

#include <config.h>
#include "translator.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <utillib.h>
#include <stdlib.h>
#ifndef YYBISON
extern void cfree(generic_ptr ptr);
#define free(p) cfree((generic_ptr) p)
#endif

#define STR_EQUAL(s1, s2) (! strcmp((s1), (s2)))
#define FLOATFMT "%0.16e"

static char dummy_string[] = NODESCRIPTION;
static unsigned int current_time_index = 0;
static char *current_drift = NULL;
static deprec *current_drift_dependencies = NULL;

static char *original_unary_functions[] = {
  "sin",
  "cos",
  "tan",
  "ceil",
  "floor",
  "atan",
  "asin",
  "acos",
  "abs",
  "fabs",
  "log",
  "log10",
  "exp",
  "sqrt",

  /* Defined in gnanslib.h */
  "sqr",
  "cube",
  "heaviside",
  
  ""};

static char *original_binary_functions[] = {
  "atan2",
  "max",
  "min",
  ""};

static char **unary_functions = original_unary_functions;
static char **binary_functions = original_binary_functions;

unsigned int 
size(char **s)
{
  return **s == '\0' ? 0 : size(s + 1) + 1;
}

static void
add_function(char **&functions, char *fname)
{
  unsigned int s = size(functions);
  char **new_functions = new char*[s+2];
  for (unsigned int i = 0; i < s; i++)
    new_functions[i] = functions[i];
  //delete(functions);
  functions = new_functions;
  functions[s] = fname;
  functions[s+1] = stringify("");
}

static void
add_unary_function(char *fname)
{
  add_function(unary_functions, fname);
}

static void
add_binary_function(char *fname)
{
  add_function(binary_functions, fname);
}

/*
 * Increments function_calls, unless the calling function is well-known.
 */
static void 
increment_function_calls(char *function_name, int number_of_args)
{
  /* Check number of args if the function is known to us */
  int i;
  for (i = 0; unary_functions[i][0]; i++)
    if (! strcmp(unary_functions[i], function_name)) {
      if (number_of_args != 1)
	gnans_error(semantic, "``%s'' called with wrong number of arguments",
		    function_name);
      return;
    }
  for (i = 0; binary_functions[i][0]; i++)
    if (! strcmp(binary_functions[i], function_name)) {
      if (number_of_args != 2)
	gnans_error(semantic, "``%s'' called with wrong number of arguments",
		    function_name);
      return;
    }

  /* Check if the name is in the symbol table */
  if (getsym(function_name) && getsym(function_name)->type != Defun)
    gnans_error(semantic,
		"`%s' can't be used as a function (forgotten `*'?)",
		function_name);

  function_calls++;
}

/* Check that we don't overfill tables of size MAXSTATES */
/* (Yes, this is bad programming...) */
static void
check_size(const unsigned int size,
	   const unsigned int max_size = SMALLTABLESIZE)
{
  if (size >= max_size) {
    gnans_error(fatal,
		"Internal limit exceeded, sorry (MAXSTATES or SMALLTABLESIZE too small).");
    if (translator_debug)
      abort();
  }
}

static bool
current_block_compare(unsigned int block)
{
  return current_block == &(times_buffer[block]);
}

#define operation stringify
%}

%union {
  double value;
  char *text;
  int keyword;
  int integer;
}

%token <value> FLOAT
%token <integer> INT
%token <text> ID
%token <text> ID8
%token <text> STRING

%token <text> COMMENT
%token C_BLOCK

%token <keyword> CONTINUOUS
%token <keyword> DISCRETE
%token <keyword> HYBRID
%token <keyword> DAS
%token <keyword> RICCATI
%token <keyword> TIME
%token <keyword> SYSTEM

%token <keyword> STATE
%token <keyword> INTEGERSTATE
%token <keyword> INTEGER	/* Note: Keyword, not an integer variable  */
%token <keyword> DOUBLE
%token <keyword> PARAMETER
%token <keyword> FUNCTION
%token <keyword> VARIABLE
%token <keyword> UNIFORM
%token <keyword> NORMAL
%token <keyword> WHITE
%token <keyword> NOISE
%token <keyword> WIENER
%token <keyword> PROCESS
%token <keyword> EVENT
%token <keyword> MATRIX
%token <keyword> VECTOR
%token <keyword> C_IDENTIFIER
%token <keyword> FILEREAD

%token <keyword> ATTIME

%right '='
%right '?' ':'
%left OR
%left AND
%left '|'
%left '&'
%left EQ NE
%left '<' '>' GE LE
%left LL GG
%left '-' '+'
%left '*' '/' '%'
%left UMINUS UPLUS '!' '~' PLUSPLUS MINUSMINUS
%right '^'			/* Exponentiation */

%type <text> string
%type <value> real;
%type <text> exp;
%type <text> formal_param;
%type <text> formal_param_list;
%type <text> multiplier;
%type <text> arglist;
%type <text> bracket_args;
%type <text> latin1id;
%type <text> fname;
%type <text> int_cast;
%type <text> double_cast;
%%
system:	        pass_thrus system_decl decls time_block time_blocks
;

decls:		/* Empty */
  		| decls decl
;

decl:		component_decl
		| time_decl
                | function_def
		| pass_thru
	        | error
;

time_blocks:	/* Empty */ 
		| time_blocks time_block
;

time_block:	time_header C_block equations
;

time_header: 	ATTIME ID ':' {
  unsigned int i;
  for (i = 0; i < number_of_times; i++) {
    if (STR_EQUAL(times_names[i].c_str(), $2))
      break;
  }
  if (i < number_of_times) {	/* Found it! */
    current_block = &(times_buffer[i]);
    current_time_index = i;
  } else
    gnans_error(semantic, "`%s' is not a declared time", $2);
}
;

C_block:	/* Empty */
  		| C_BLOCK
;

formal_param:   ID              { putsym($1, Formal_Parameter);
				  $$ = stringify("double %s", $1); }
;

formal_param_list: 
		formal_param ',' formal_param ',' formal_param
                    { $$ = stringify("%s, %s, %s", $1, $3, $5); }
		| formal_param ',' formal_param_list
                    { $$ = stringify("%s,%s", $1,$3); }
;
               
function_def:	unary_function_def
                | binary_function_def
                | general_function_def
;

unary_function_def: ID '(' formal_param ')' '=' exp ';'
{
  fprintf(zzout, "static double %s(%s){return %s;}\n", $1, $3, $6);
  fprintf(zzout, "extern void NeVeR_CaLlEd_%s(){%s(0);}\n", $1, $1);
  symtab_delete_type(Formal_Parameter);
  current_dependencies = NULL;
  put_unique_sym($1, Defun);
  add_unary_function($1);
}
;

binary_function_def: ID '(' formal_param ',' formal_param')' '=' exp ';'
{
  fprintf(zzout, "static double %s(%s,%s){return %s;}\n", $1, $3, $5, $8);
  fprintf(zzout, "extern void NeVeR_CaLlEd_%s(){%s(0,0);}\n", $1, $1);
  symtab_delete_type(Formal_Parameter);
  current_dependencies = NULL;
  put_unique_sym($1, Defun);
  add_binary_function($1);
}
;

general_function_def:  ID '(' formal_param_list ')' '=' exp ';'
{
  fprintf(zzout, "static double %s(%s){return %s;}\n", $1, $3, $6);
  fprintf(zzout, "extern void *NeVeR_CaLlEd_%s(){ return (void*)%s;}\n",
	    $1, $1);
  symtab_delete_type(Formal_Parameter);
  current_dependencies = NULL;
  put_unique_sym($1, Defun);
}
;

equations:	/* Empty */
  		| equations equation
;

equation:	diff_eqn
		| stoch_diff_eqn
		| difference_eqn
		| alg_eqn
                | ricc_eqn
                | implicit_eqn
		| component_decl
		| pass_thru
	        | error
;

ricc_eqn:	'd' '(' ID ')' '=' RICCATI '('
                                   ID ','
                                   ID ','
                                   ID ','
                                   ID ','
                                   ID ')' ';'
{
  if (system_type != riccati)
    gnans_error(semantic, "Riccati equations only allowed in Riccati systems");
  symrec *thestate = getsym($3);
  if (thestate) {
    if (thestate->type != State_matrix)
      gnans_error(semantic,
		  "`%s' is not a state matrix; "
		  "it cannot have a differential Riccati equation",
		  $3);
    if (thestate->equation)
      gnans_error(semantic, "Repeated equation for `d(%s)'", $3);
    
    if (!STR_EQUAL($3, $8))
      gnans_error(semantic, "First argument of RICCATI must be the state matrix");
    if (!getsym($10))
      gnans_error(semantic, "`%s' is not defined", $10);
    if (!getsym($12))
      gnans_error(semantic, "`%s' is not defined", $12);
    if (!getsym($14))
      gnans_error(semantic, "`%s' is not defined", $14);
    if (!getsym($16))
      gnans_error(semantic, "`%s' is not defined", $16);
  
    if (!((getsym($10)->type == Parameter_matrix) &&
	  (getsym($12)->type == Parameter_matrix) &&
	  (getsym($14)->type == Parameter_matrix) &&
	  (getsym($16)->type == Parameter_matrix)))
      gnans_error(semantic,
		  "Arguments 2, 3, 4, and 5 of RICCATI "
		  "must be parameter matrices");
  
    if (current_block_compare(STARTINDEX))
      gnans_error(semantic, "Riccati equations not allowed in AT TIME START");
    if (current_block_compare(STOPINDEX))
      gnans_error(semantic, "Riccati equations not allowed in AT TIME STOP");
    if (current_block_compare(STOREINDEX))
      gnans_error(semantic, "Riccati equations not allowed in AT TIME STORE");
    if (current_block_compare(ACCEPTEDINDEX))
      gnans_error(semantic, "Riccati equations not allowed in AT TIME ACCEPTED");
    thestate->a_priori_timeindex = current_time_index;

    thestate->equation = stringify("//RICCATI(%s,%s,%s,%s,%s)",
				   $8, $10, $12, $14, $16);
    thestate->dependencies = current_dependencies;
    thestate->line_no_of_equation = lineno;
    current_dependencies = NULL;
    riccati_data.A21 = getsym($10);
    riccati_data.A22 = getsym($12);
    riccati_data.A11 = getsym($14);
    riccati_data.A12 = getsym($16);
  }
  else
    gnans_error(semantic, "`%s' is not defined", $3);
}
;

diff_eqn:	'd' '(' ID ')' '=' exp ';'
{
  symrec *thestate, *thewiener, *sym;
  unsigned int i;
  if (system_type == discrete || system_type == riccati)
    gnans_error(semantic,
		"Diffential equations are not allowed in discrete time systems");
  thestate = getsym($3);
  if (thestate) {
    if (thestate->type != State)
      gnans_error(semantic,
		  "`%s' is not a state; it cannot have a differential equation",
		  $3);
    if (thestate->equation)
      gnans_error(semantic, "Repeated equation for `d(%s)'", $3);
    if (current_block_compare(STARTINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME START");
    if (current_block_compare(STOPINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME STOP");
    if (current_block_compare(STOREINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME STORE");
    if (current_block_compare(ACCEPTEDINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME ACCEPTED");
    thestate->a_priori_timeindex = current_time_index;
    thestate->equation = $6;
    thestate->dependencies = current_dependencies;
    thestate->line_no_of_equation = lineno;
    current_dependencies = NULL;

    for (i = 0; i < number_of_wieners; i++) {
      thewiener = getsym((char*)wieners[i] + 2);	/* Zap the "d_" */
      sym = put_unique_sym(stringify("%s_d_%s", thestate->name,
				     thewiener->name), Wiener_coeff);
      sym->equation = stringify("0");
      sym->dependencies = NULL;
      sym->line_no_of_equation = lineno;
    }
  }
  else
    gnans_error(semantic, "`%s' is not defined", $3);
}
;					     

difference_eqn:	'q' '(' ID ')' '=' exp ';'
{
  symrec *thestate;
  if (system_type != discrete)
    gnans_error(semantic, "Diffence equations only allowed in discrete time systems");
  thestate = getsym($3);
  if (thestate) {
    if (thestate->type != State)
      gnans_error(semantic,
		  "`%s' is not a state; it cannot have a difference equation",
		  $3);
    if (thestate->equation)
      gnans_error(semantic, "Repeated equation for `d(%s)'", $3);
    if (current_block_compare(STARTINDEX))
      gnans_error(semantic, "Difference equations not allowed in AT TIME START");
    if (current_block_compare(STOPINDEX))
      gnans_error(semantic, "Difference equations not allowed in AT TIME STOP");
    if (current_block_compare(STOREINDEX))
      gnans_error(semantic, "Difference equations not allowed in AT TIME STORE");
    if (current_block_compare(ACCEPTEDINDEX))
      gnans_error(semantic, "Difference equations not allowed in AT TIME ACCEPTED");
    thestate->a_priori_timeindex = current_time_index;
    thestate->equation = $6;
    thestate->dependencies = current_dependencies;
    thestate->line_no_of_equation = lineno;
  }
  else
    gnans_error(semantic, "`%s' is not defined", $3);
  current_dependencies = NULL;
}
;

alg_eqn:	ID '=' exp ';' {
  symrec *thevariable;
  if (system_type == riccati)
    gnans_error(semantic, "Algebraic equations not allowed in Riccati systems");
  thevariable = getsym($1);
  if (thevariable) {
    if (current_block_compare(ACCEPTEDINDEX))
      gnans_error(semantic, "Equations not allowed in AT TIME ACCEPTED");
    if (current_block_compare(STOREINDEX))
      gnans_error(semantic, "Equations not allowed in AT TIME STORE");
    if (current_block_compare(STOPINDEX))
      gnans_error(semantic, "Equations not allowed in AT TIME STOP");
    if (thevariable->type == State)
      gnans_error(semantic, "`%s' can only have a dynamic equation", $1);
    if ((thevariable->type != Variable) && (thevariable->type != Function))
      gnans_error(semantic,
		  "`%s' cannot be have an equation (only variables and functions)", $1);
    if (thevariable->equation)
      gnans_error(semantic, "Repeated equation for `%s'", $1);
    thevariable->a_priori_timeindex = current_time_index;
    thevariable->equation = $3;
    thevariable->dependencies = current_dependencies;
    thevariable->line_no_of_equation = lineno;
  }
  else
    gnans_error(semantic, "`%s' is not defined", $1);
  current_dependencies = NULL;
} 
;

implicit_eqn:   INT '=' exp ';' {
}
;

stoch_diff_eqn:	'd' '(' ID ')' '=' diffusion diffusions ';'
{
  symrec *thestate, *thewiener, *sym;
  unsigned int i;
  if (system_type == discrete)
    gnans_error(semantic,
		"Diffential equations are not allowed in discrete time systems");
  thestate = getsym($3);
  if (thestate) {
    if (thestate->type != State)
      gnans_error(semantic,
		  "`%s' is not a state; it cannot have a differential equation", $3);
    if (thestate->equation)
      gnans_error(semantic, "Repeated equation for `d(%s)'", $3);
    if (current_block_compare(STARTINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME START");
    if (current_block_compare(STOPINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME STOP");
    if (current_block_compare(STOREINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME STORE");
    if (current_block_compare(ACCEPTEDINDEX))
      gnans_error(semantic, "Differential equations not allowed in AT TIME ACCEPTED");
    thestate->a_priori_timeindex = current_time_index;
    thestate->equation = current_drift ? current_drift : stringify("0");
    current_drift = NULL;
    thestate->dependencies = current_drift_dependencies;
    current_drift_dependencies = NULL;
    thestate->line_no_of_equation = lineno;
    current_dependencies = NULL;

    for (i = 0; i < number_of_wieners; i++) {
      thewiener = getsym((char*)wieners[i] + 2);	/* Zap the "d_" */
      sym = put_unique_sym(stringify("%s_d_%s", thestate->name,
				     thewiener->name), Wiener_coeff);
      sym->equation = thewiener->equation ? thewiener->equation : stringify("0");
      thewiener->equation = NULL;
      sym->dependencies = thewiener->dependencies;
      thewiener->dependencies = NULL;
      sym->line_no_of_equation = lineno;
    }
  }
  else
    gnans_error(semantic, "`%s' is not defined", $3);
}
;					     

diffusions:	/* Empty */
		| diffusions '+' diffusion
		| diffusions minus_diffusion
;

diffusion:	'0'	{ current_drift = stringify("%s", "0");
			  current_drift_dependencies = NULL;
			  current_dependencies = NULL;}
		| multiplier 'd' '(' ID ')'
{
  unsigned int i;
  if (time_variable_name ==  $4) {
    /* This is the drift term */

    /* Barf on several drift vector fields */
    if (current_drift)
      gnans_error(semantic, "Only one drift vector field is allowed.");
    current_drift = stringify("%s", $1);
    current_drift_dependencies = current_dependencies;
    current_dependencies = NULL;
  } else {
    for (i = 0; i < number_of_wieners; i++)
      if (STR_EQUAL(wieners[i], stringify("d_%s", $4)))
	break;
    if (i == number_of_wieners)
      gnans_error(semantic, "``d(%s)'' doesn't make sense here", $4);
    else {
      symrec *the_wiener;
      the_wiener = getsym($4);	/* WILL succeed, don't check */
      if (the_wiener->equation)
	gnans_error(semantic, "Only one drift vector field for %s allowed.",
		    the_wiener->name);
      the_wiener->equation = stringify("%s", $1);
      the_wiener->dependencies = current_dependencies;
      current_dependencies = NULL;
    }
  }
}
;

minus_diffusion:  '-' multiplier 'd' '(' ID ')'
{
  unsigned int i;
  if (time_variable_name == $5) {
    /* This is the drift term */

    /* Barf on several drift vector fields */
    if (current_drift)
      gnans_error(semantic, "Only one drift vector field is allowed.");
    current_drift = stringify("-%s", $2);
    current_drift_dependencies = current_dependencies;
    current_dependencies = NULL;
  } else {
    for (i = 0; i < number_of_wieners; i++)
      if (STR_EQUAL(wieners[i], stringify("d_%s", $5)))
	break;
    if (i == number_of_wieners)
      gnans_error(semantic, "``d(%s)'' doesn't make sense here", $5);
    else {
      symrec *the_wiener;
      the_wiener = getsym($5);	/* WILL succeed, don't check */
      if (the_wiener->equation)
	gnans_error(semantic, "Only one drift vector field for %s allowed.",
		    the_wiener->name);
      the_wiener->equation = stringify("-%s", $2);
      the_wiener->dependencies = current_dependencies;
      current_dependencies = NULL;
    }
  }
}
;

multiplier:	/* Empty */	{ $$ = stringify("1"); }
		| exp '*'	{ $$ = $1; }
;

exp:		FLOAT		{ $$ = stringify(FLOATFMT,$1); }
		| INT		{ $$ = stringify("%d.0", $1); }
	        | ID            { add_dependency($1); $$ = $1; }
	        | 'q' '(' ID ')' { /*!!!!*/add_dependency($3);
					   $$ = stringify("q(%s)", $3); }
	        | int_cast exp %prec UMINUS 
				{ $$ = stringify("%s %s", $1, $2); }
	        | double_cast exp %prec UMINUS 
				{ $$ = operation("%s %s", $1, $2); }
	        | fname '(' ')' { $$ = operation("%s()", $1);
					 increment_function_calls($1, 0); }
	        | fname '(' exp ')' { $$ = operation("%s(%s)", $1, $3);
					 increment_function_calls($1, 1); }
	        | fname '(' exp ',' exp ')'   
				{ $$ = operation("%s(%s,%s)", $1, $3, $5);
                                  increment_function_calls($1, 2); }
	        | fname '(' exp ',' exp ',' arglist ')'
				{ $$ = operation("%s(%s,%s,%s)", $1, $3, $5, $7);
					 increment_function_calls($1, 3); }
	        | ID bracket_args
				{ $$ = operation("%s%s", $1, $2); }
	        | exp '+' exp   { $$ = operation("%s+%s", $1, $3); }
        	| exp '-' exp   { $$ = operation("%s-%s", $1, $3); }
	        | exp '*' exp   { $$ = operation("%s*%s", $1, $3);}
        	| exp '/' exp   { $$ = operation("%s/%s", $1, $3); }
	        | exp '%' exp   { $$ = operation("fmod(%s,%s)", $1, $3); }
        	| PLUSPLUS ID   { $$ = operation("++%s", $2); }
	        | MINUSMINUS ID { $$ = operation("--%s", $2); } 
        	| ID PLUSPLUS   { $$ = operation("%s++", $1); }
	        | ID MINUSMINUS { $$ = operation("%s--", $1); }
        	| '-' exp  %prec UMINUS { $$ = operation("-%s", $2); }
        	| '+' exp  %prec UPLUS  { $$ = operation("-%s", $2); }
        	| '!' exp        { $$ = operation("!%s", $2); }
        	| '~' exp        { $$ = operation("~%s", $2); }
        	| exp '^' exp    { $$ = operation("pow((double)%s,(double)%s)",
						  $1, $3); }
        	| '(' exp ')'    { $$ = operation("(%s)", $2); }
		| exp LL exp     { $$ = operation("%s<<%s", $1, $3); }
	        | exp GG exp     { $$ = operation("%s>>%s", $1, $3); }
	        | exp EQ exp     { $$ = operation("%s==%s", $1, $3); } 
        	| exp NE exp     { $$ = operation("%s!=%s", $1, $3); } 
	        | exp LE exp     { $$ = operation("%s<=%s", $1, $3); } 
        	| exp GE exp     { $$ = operation("%s>=%s", $1, $3); }
	        | exp '<' exp    { $$ = operation("%s<%s", $1, $3); }
        	| exp '>' exp    { $$ = operation("%s>%s", $1, $3); }
	        | exp '&' exp    { $$ = operation("%s&%s", $1, $3); }
        	| exp '|' exp    { $$ = operation("%s|%s", $1, $3); }
	        | exp AND exp    { $$ = operation("%s&&%s", $1, $3); }
        	| exp OR exp     { $$ = operation("%s||%s", $1, $3); }
        	| exp '?' exp ':' exp { $$ = operation("%s?%s:%s", $1, $3, $5); }
;

fname:		ID		{ $$ = $1; }
		| fname '.' ID	{ $$ = stringify("%s.%s", $1, $3); }
;

arglist: 	exp		{ $$ = stringify("%s", $1); }
		| arglist ',' exp
				{ $$ = stringify("%s,%s", $1, $3); }
;

int_cast:	'(' INTEGER ')'	{ $$ = stringify("(int)"); }
;

double_cast:	'(' DOUBLE ')'	{ $$ = stringify("(double)"); }
;

bracket_args:   '[' exp ']'	{ $$ = stringify("[(int)%s]", $2); }
  		| bracket_args '[' exp ']'
       				{ $$ = stringify("%s[(int)%s]", $1, $3); }
;

time_decl:	TIME ID ';' {
  check_size(number_of_times);
  put_unique_sym($2, Time_type);
  times_buffer[number_of_times] = "";
  time_variable_name = $2;
  times_names[number_of_times++] = $2;
}
;

system_decl:
		  continuous_system
		| discrete_system
		| hybrid_system
		| das_system
		| riccati_system
;

latin1id:	ID
  		| ID8	{ $$ = $1;}
;

continuous_system:
		CONTINUOUS TIME SYSTEM latin1id ';' {
		  system_type = continuous;
		  system_name = $4; }
;

discrete_system:
		DISCRETE TIME SYSTEM latin1id ';' {
		  system_type = discrete; 
		  system_name = $4; }
;

hybrid_system:
		HYBRID TIME SYSTEM latin1id ';' { system_type = hybrid; 
					    system_name = $4; }
;

das_system:
		DAS SYSTEM latin1id ';' { system_type = das; 
					  system_name = $3; }
;

riccati_system:
		RICCATI SYSTEM latin1id ';' {system_type = riccati; 
					     system_name = $3; }
;

component_decl:
  		  state_decl
  		| int_state_decl
		| par_decl
		| func_decl
  		| var_decl
		| u_noise_decl
  		| n_noise_decl
  		| wiener_decl
  		| id_decl
;

state_decl:
  		scalar_states_decl
  		| vector_state_decl
		| matrix_state_decl	{}
;

scalar_states_decl:      STATE scalar_states ';' {}
;

scalar_states:           scalar_state_decl
                | scalar_state_decl ',' scalar_states
;

scalar_state_decl:	ID '=' real string {
  check_size(number_of_states, MAXSTATES);
  put_unique_sym($1, State);
  states[number_of_states] = $1;
  states[number_of_states].initial = $3;
  states[number_of_states++].documentation = stringify("%s", $4);
  fprintf(zzout, "static UserParameter gnans_state_%s;\n", $1);
}
;

vector_state_decl:	STATE VECTOR ID '=' 
                          FILEREAD '(' ID ',' string ',' string ')'  string ';'
                  { gnans_error(semantic, "VECTOR not yet implemented"); }
;

matrix_state_decl:	STATE MATRIX ID '='
                          FILEREAD '(' ID ',' string ',' string ')' string ';'
                  { check_size(number_of_state_matrices);
		    if (system_type != riccati)
		      gnans_error(semantic, "MATRIX not yet implemented");
		    if (number_of_state_matrices)
		      gnans_error(semantic, "Only one STATE MATRIX allowed");
		    put_unique_sym($3, State_matrix);
		    state_matrices[number_of_state_matrices] = $3;
		    state_matrices[number_of_state_matrices].filename
		      = stringify("%s", $9);
		    state_matrices[number_of_state_matrices].external_name
		      = stringify("%s", $11);
		    //states[number_of_states].initial = 0;
		    state_matrices[number_of_state_matrices++].documentation
		      = stringify("%s", $13);
		    if (!STR_EQUAL($7, "matlab"))
		      gnans_error(semantic,
				  "Only Matlab format presently implemented");
		    fprintf(zzout, 
			    "static user_matrix gnans_state_matrix_%s;\n", $3);
}
;

int_state_decl:	INTEGERSTATE int_states ';' {}
;

int_states:     int_state
                | int_state ',' int_states
;

int_state:	ID '=' INT string
{
  /* Presently, we treat the integer states really stepmotherly */
  put_unique_sym($1, Integer_State);
  c_declarations += "int " + std::string($1) + " = " + number_String($3)
    + "; /* " + $4 + " */\n" ;
  times_buffer[STARTINDEX] += stringify("  %s = %d;\n", $1, $3);
}
;

par_decl:	  scalar_par_decl
                | vector_par_decl
  		| matrix_par_decl {}
;

scalar_par_decl:  PARAMETER scalar_pars ';' {}
;

scalar_pars:      scalar_par
		| scalar_par ',' scalar_pars
;

scalar_par:	ID '=' real string {
  check_size(number_of_parameters, MAXSTATES);
  put_unique_sym($1, Parameter);
  parameters[number_of_parameters] = $1;
  parameters[number_of_parameters].initial = $3;
  parameters[number_of_parameters++].documentation = stringify("%s", $4);
  
  fprintf(zzout, "static UserParameter gnans_parameter_%s;\n", $1);
}
;

vector_par_decl:	PARAMETER VECTOR ID '=' 
                         FILEREAD '(' ID ',' string ',' string ')' string ';' 
                { gnans_error(semantic, "VECTOR not yet implemented"); }
;

matrix_par_decl:	PARAMETER MATRIX ID '=' 
                         FILEREAD '(' ID ',' string ',' string ')' string ';'
{
  check_size(number_of_parameter_matrices);
  if (system_type != riccati)
    gnans_error(semantic, "MATRIX not yet implemented");
  put_unique_sym($3, Parameter_matrix);
  parameter_matrices[number_of_parameter_matrices] = $3;
  parameter_matrices[number_of_parameter_matrices].filename
    = stringify("%s", $9);
  parameter_matrices[number_of_parameter_matrices].external_name
    = stringify("%s", $11);
  parameter_matrices[number_of_parameter_matrices++].documentation
    = stringify("%s", $13);
  if (!STR_EQUAL($7, "matlab"))
    gnans_error(semantic, "Only Matlab format presently implemented");
  fprintf(zzout, "static user_matrix gnans_parameter_matrix_%s;\n", $3);
}
;

func_decl:      FUNCTION functs_decl ';' {}
;

functs_decl:    func
                | func ',' functs_decl
;

func:	ID string {
  check_size(number_of_functions, MAXSTATES);
  put_unique_sym($1, Function);
  functions[number_of_functions] = $1;
  functions[number_of_functions++].documentation = stringify("%s", $2);
  fprintf(zzout, "static UserParameter gnans_function_%s;\n", $1);
}
;

var_decl:	VARIABLE vars_decls ';' {}
;

vars_decls:      var_dec
                | var_dec ',' vars_decls
;

var_dec:	ID {
  check_size(number_of_variables, MAXSTATES);
  put_unique_sym($1, Variable);
  variables[number_of_variables++] = $1;
  fprintf(zzout, "static double gnans_variable_%s;\n", $1);
}
;

id_decl:	C_IDENTIFIER id_decls ';' {}
;

id_decls:      id
               | id ',' id_decls
;

id:	ID {
  put_unique_sym($1, C_Identifier);
}
;

u_noise_decl:	UNIFORM WHITE NOISE u_noises_decl ';' {}
;

u_noises_decl:  u_noise
                | u_noise ',' u_noises_decl
;

u_noise:	ID '=' INT string {
  check_size(number_of_randoms, MAXSTATES);
  check_size(number_of_uniforms, MAXSTATES);
  put_unique_sym($1, Uniform_Noise);
  randoms[number_of_randoms] = $1;
  randoms[number_of_randoms].initial = $3;
  randoms[number_of_randoms++].documentation = stringify("%s", $4);
  uniforms[number_of_uniforms] = $1;
  uniforms[number_of_uniforms].initial = $3;
  uniforms[number_of_uniforms++].documentation = stringify("%s", $4);
  fprintf(zzout, "static UserParameter gnans_random_%s;\n", $1);
};

n_noise_decl:	NORMAL WHITE NOISE n_noises_decl ';' {}
;

n_noises_decl:  n_noise
                | n_noise ',' n_noises_decl
;

n_noise:	ID '=' INT string {
  check_size(number_of_randoms, MAXSTATES);
  check_size(number_of_normals, MAXSTATES);
  put_unique_sym($1, Normal_Noise);
  randoms[number_of_randoms] = $1;
  randoms[number_of_randoms].initial = $3;
  randoms[number_of_randoms++].documentation = stringify("%s", $4);
  normals[number_of_normals] = $1;
  normals[number_of_normals].initial = $3;
  normals[number_of_normals++].documentation = stringify("%s", $4);
  fprintf(zzout, "static UserParameter gnans_random_%s;\n", $1);
};

wiener_decl:    WIENER PROCESS wieners_decl ';' {}
;

wieners_decl:   wiener
                | wiener ',' wieners_decl
;

wiener:    ID '=' INT string {
  check_size(number_of_randoms, MAXSTATES);
  check_size(number_of_wieners, MAXSTATES);
  put_unique_sym($1, Wiener_Process);
  std::string name = std::string("d_") + $1;
  randoms[number_of_randoms] = name;
  randoms[number_of_randoms].initial = $3;
  randoms[number_of_randoms++].documentation = stringify("%s", $4);
  wieners[number_of_wieners] = name;
  wieners[number_of_wieners].initial = $3;
  wieners[number_of_wieners++].documentation = stringify("%s", $4);
  fprintf(zzout, "static UserParameter gnans_random_%s;\n", name.c_str());
}
;

string:		  /* Empty */ { $$ = dummy_string; }
  		| STRING {
                            /* Zap last `"' */
		            zztext[strlen((const char *) zztext) - 1] = '\0';
			    /* Zap first `"`, and make a copy */
			    $$ = stringify("%s", (char *) zztext + 1);
			  }
;

real:		FLOAT		{ $$ = $1; }
		| '-' FLOAT	{ $$ = -$2; }
		| INT		{ $$ = (double) $1; }
		| '-' INT	{ $$ = - (double) $2; }
;
pass_thrus:	/* Empty */
		| pass_thrus pass_thru
;
pass_thru:	';'	     /* Allow spurious `;' */
		| COMMENT { gnans_error(impossible, "");
#if 0
#ifdef YYBISON
			    if (lineno)	/* To prevent complains... */
			      goto zzerrlab1;
			    else
			      goto zzbackup;
#else
			    if (lineno)	/* To prevent complains... */
			      goto zzerrlab;
			    else
			      goto zznewstate;
#endif
#endif
}
;
%%
